from .abstract_cub_bayes_ld_g import AbstractCubBayesLDG
from ..discrete_distribution import DigitalNetB2
from ..integrand import Keister,BoxIntegral,Genz,SensitivityIndices
from ..fast_transform import fwht,omega_fwht
from ..util import MaxSamplesWarning, ParameterError, ParameterWarning, NotYetImplemented
import ctypes
import numpy as np
from time import time
import warnings


class CubQMCBayesNetG(AbstractCubBayesLDG):
    r"""
    Quasi-Monte Carlo stopping criterion using fast Bayesian cubature and digital nets
    with guarantees for Gaussian processes having certain digitally shift invariant kernels.

    Examples:
        >>> k = Keister(DigitalNetB2(2, seed=123456789))
        >>> sc = CubQMCBayesNetG(k,abs_tol=5e-3)
        >>> solution,data = sc.integrate()
        >>> data
        Data (Data)
            solution        1.809
            comb_bound_low  1.807
            comb_bound_high 1.810
            comb_bound_diff 0.003
            comb_flags      1
            n_total         2^(10)
            n               2^(10)
            time_integrate  ...
        CubQMCBayesNetG (AbstractStoppingCriterion)
            abs_tol         0.005
            rel_tol         0
            n_init          2^(8)
            n_limit         2^(22)
            order           1
        Keister (AbstractIntegrand)
        Gaussian (AbstractTrueMeasure)
            mean            0
            covariance      2^(-1)
            decomp_type     PCA
        DigitalNetB2 (AbstractLDDiscreteDistribution)
            d               2^(1)
            replications    1
            randomize       LMS DS
            gen_mats_source joe_kuo.6.21201.txt
            order           RADICAL INVERSE
            t               63
            alpha           1
            n_limit         2^(32)
            entropy         123456789
    
        Vector outputs
        
        >>> f = BoxIntegral(DigitalNetB2(3,seed=7),s=[-1,1])
        >>> abs_tol = 1e-2
        >>> sc = CubQMCBayesNetG(f,abs_tol=abs_tol,rel_tol=0)
        >>> solution,data = sc.integrate()
        >>> solution
        array([1.18750491, 0.96076395])
        >>> data
        Data (Data)
            solution        [1.188 0.961]
            comb_bound_low  [1.18 0.96]
            comb_bound_high [1.195 0.962]
            comb_bound_diff [0.014 0.002]
            comb_flags      [ True  True]
            n_total         2^(11)
            n               [2048  256]
            time_integrate  ...
        CubQMCBayesNetG (AbstractStoppingCriterion)
            abs_tol         0.010
            rel_tol         0
            n_init          2^(8)
            n_limit         2^(22)
            order           1
        BoxIntegral (AbstractIntegrand)
            s               [-1  1]
        Uniform (AbstractTrueMeasure)
            lower_bound     0
            upper_bound     1
        DigitalNetB2 (AbstractLDDiscreteDistribution)
            d               3
            replications    1
            randomize       LMS DS
            gen_mats_source joe_kuo.6.21201.txt
            order           RADICAL INVERSE
            t               63
            alpha           1
            n_limit         2^(32)
            entropy         7
        >>> sol3neg1 = -np.pi/4-1/2*np.log(2)+np.log(5+3*np.sqrt(3))
        >>> sol31 = np.sqrt(3)/4+1/2*np.log(2+np.sqrt(3))-np.pi/24
        >>> true_value = np.array([sol3neg1,sol31])
        >>> assert (abs(true_value-solution)<abs_tol).all()

        Sensitivity indices 

        >>> function = Genz(DigitalNetB2(3,seed=7))
        >>> integrand = SensitivityIndices(function)
        >>> sc = CubQMCBayesNetG(integrand,abs_tol=5e-2,rel_tol=0)
        >>> solution,data = sc.integrate()
        >>> data
        Data (Data)
            solution        [[0.009 0.194 0.657]
                             [0.036 0.312 0.783]]
            comb_bound_low  [[0.007 0.178 0.636]
                             [0.033 0.297 0.762]]
            comb_bound_high [[0.012 0.209 0.679]
                             [0.038 0.327 0.804]]
            comb_bound_diff [[0.005 0.031 0.043]
                             [0.004 0.03  0.042]]
            comb_flags      [[ True  True  True]
                             [ True  True  True]]
            n_total         2^(8)
            n               [[[256 256 256]
                              [256 256 256]
                              [256 256 256]]
        <BLANKLINE>
                             [[256 256 256]
                              [256 256 256]
                              [256 256 256]]]
            time_integrate  ...
        CubQMCBayesNetG (AbstractStoppingCriterion)
            abs_tol         0.050
            rel_tol         0
            n_init          2^(8)
            n_limit         2^(22)
            order           1
        SensitivityIndices (AbstractIntegrand)
            indices         [[ True False False]
                             [False  True False]
                             [False False  True]]
        Uniform (AbstractTrueMeasure)
            lower_bound     0
            upper_bound     1
        DigitalNetB2 (AbstractLDDiscreteDistribution)
            d               3
            replications    1
            randomize       LMS DS
            gen_mats_source joe_kuo.6.21201.txt
            order           RADICAL INVERSE
            t               63
            alpha           1
            n_limit         2^(32)
            entropy         7
        
    **References:**

    1.  Jagadeeswaran, Rathinavel, and Fred J. Hickernell.  
        "Fast automatic Bayesian cubature using Sobol’sampling."  
        Advances in Modeling and Simulation: Festschrift for Pierre L'Ecuyer.  
        Springer International Publishing, 2022. 301-318.
    
    2.  Jagadeeswaran Rathinavel,  
        Fast automatic Bayesian cubature using matching kernels and designs,  
        PhD thesis, Illinois Institute of Technology, 2019.

    3.  Sou-Cheng T. Choi, Yuhan Ding, Fred J. Hickernell, Lan Jiang, Lluis Antoni Jimenez Rugama,  
        Da Li, Jagadeeswaran Rathinavel, Xin Tong, Kan Zhang, Yizhi Zhang, and Xuan Zhou,  
        GAIL: Guaranteed Automatic Integration Library (Version 2.3) [MATLAB Software], 2019.  
        [http://gailgithub.github.io/GAIL_Dev/](http://gailgithub.github.io/GAIL_Dev/).  
        [https://github.com/GailGithub/GAIL_Dev/blob/master/Algorithms/IntegrationExpectation/cubBayesNet_g.m](https://github.com/GailGithub/GAIL_Dev/blob/master/Algorithms/IntegrationExpectation/cubBayesNet_g.m).
    """

    def __init__(self, 
                 integrand, 
                 abs_tol = 1e-2, 
                 rel_tol = 0,
                 n_init = 2**8, 
                 n_limit = 2**22, 
                 error_fun = "EITHER", 
                 alpha = 0.01,
                 errbd_type="MLE",
                 ):
        r"""
        Args:
            integrand (AbstractIntegrand): The integrand.
            abs_tol (np.ndarray): Absolute error tolerance.
            rel_tol (np.ndarray): Relative error tolerance.
            n_init (int): Initial number of samples. 
            n_limit (int): Maximum number of samples.
            error_fun (Union[str,callable]): Function mapping the approximate solution, absolute error tolerance, and relative error tolerance to the current error bound.

                - `'EITHER'`, the default, requires the approximation error must be below either the absolue *or* relative tolerance.
                    Equivalent to setting
                    ```python
                    error_fun = lambda sv,abs_tol,rel_tol: np.maximum(abs_tol,abs(sv)*rel_tol)
                    ```
                - `'BOTH'` requires the approximation error to be below both the absolue *and* relative tolerance. 
                    Equivalent to setting
                    ```python
                    error_fun = lambda sv,abs_tol,rel_tol: np.minimum(abs_tol,abs(sv)*rel_tol)
                    ```
            alpha (np.ndarray): Uncertainty level in $(0,1)$. 
            errbd_type (str): Options are 
                
                - `'MLE'`: Marginal Log Likelihood. 
                - `'GCV'`: Generalized Cross Validation. 
                - `'FULL'`: Full Bayes.
        """
        super(CubQMCBayesNetG, self).__init__(integrand, ft=fwht, omega=omega_fwht,
                                           ptransform=None,
                                           allowed_distribs=[DigitalNetB2],
                                           kernel=self._shift_inv_kernel_digital,
                                           abs_tol=abs_tol, rel_tol=rel_tol,
                                           n_init=n_init, n_limit=n_limit, alpha=alpha, error_fun=error_fun, errbd_type=errbd_type)
        self.order = 1  # Currently supports only order=1
        # private properties
        # Full Bayes - assumes m and s^2 as hyperparameters
        # GCV - Generalized cross validation
        self.kernType = 1  # Type-1:
        self._xfullundtype = np.uint64
        if self.discrete_distrib.order!='RADICAL INVERSE':
            raise ParameterError("CubQMCNet_g requires DigitalNetB2 with 'RADICAL INVERSE' order")


    # Digitally shift invariant kernel
    # C1 : first row of the covariance matrix
    # Lambda : eigen values of the covariance matrix
    # Lambda_ring = fwht(C1 - 1)

    def _shift_inv_kernel_digital(self, xun, order, a, avoid_cancel_error, kern_type, debug_enable):
        kernel_func = CubQMCBayesNetG.BuildKernelFunc(order)
        const_mult = 1 / 10

        if avoid_cancel_error:
            # Computes C1m1 = C1 - 1
            # C1_new = 1 + C1m1 indirectly computed in the process
            (vec_C1m1, C1_alt) = self.kernel_t(a * const_mult, kernel_func(xun))
            lambda_factor = max(abs(vec_C1m1))
            C1_alt = C1_alt / lambda_factor
            vec_C1m1 = vec_C1m1 / lambda_factor

            # eigenvalues must be real : Symmetric pos definite Kernel
            vec_lambda_ring = np.real(fwht(vec_C1m1.copy())*np.sqrt(vec_C1m1.shape[-1]))

            vec_lambda = vec_lambda_ring.copy()
            vec_lambda[0] = vec_lambda_ring[0] + len(vec_lambda_ring) / lambda_factor

            if debug_enable:
                # eigenvalues must be real : Symmetric pos definite Kernel
                vec_lambda_direct = np.real(
                    np.array(fwht(C1_alt)*np.sqrt(C1_alt.shape[-1]), dtype=float))  # Note: fwht output not normalized
                if sum(abs(vec_lambda_direct - vec_lambda)) > 1:
                    print('Possible error: check vec_lambda_ring computation')
        else:
            # direct approach to compute first row of the kernel Gram matrix
            vec_C1 = np.prod(1 + a * const_mult * kernel_func(xun), 2)
            # eigenvalues must be real : Symmetric pos definite Kernel
            vec_lambda = np.real(fwht(vec_C1)*np.sqrt(vec_C1.shape[-1]))
            vec_lambda_ring = 0
            lambda_factor = 1

        return vec_lambda, vec_lambda_ring, lambda_factor

    # Builds High order walsh kernel function
    @staticmethod
    def BuildKernelFunc(order):
        # a1 = @(x)(-np.floor(np.log2(x)))
        def a1(x):
            out = -np.floor(np.log2(x + np.finfo(float).eps))
            out[x == 0] = 0  # a1 is zero when x is zero
            return out

        # t1 = @(x)(2.^(-a1(x)))
        def t1(x):
            out = 2 ** (-a1(x))
            out[x == 0] = 0  # t1 is zero when x is zero
            return out

        # t2 = @(x)(2.^(-2*a1(x)))
        def t2(x):
            out = (2 ** (-2 * a1(x)))
            out[x == 0] = 0  # t2 is zero when x is zero
            return out

        s1 = lambda x: (1 - 2 * x)
        s2 = lambda x: (1 / 3 - 2 * (1 - x) * x)
        ts2 = lambda x: ((1 - 5 * t1(x)) / 2 - (a1(x) - 2) * x)
        ts3 = lambda x: ((1 - 43 * t2(x)) / 18 + (5 * t1(x) - 1) * x + (a1(x) - 2) * x ** 2)

        if order == 1:
            kernFunc = lambda x: (6 * ((1 / 6) - 2 ** (np.floor(np.log2(x + np.finfo(float).eps)) - 1)))
        elif order == 2:
            omega2_1D = lambda x: (s1(x) + ts2(x))
            kernFunc = omega2_1D
        elif order == 3:
            omega3_1D = lambda x: (s1(x) + s2(x) + ts3(x))
            kernFunc = omega3_1D
        else:
            NotYetImplemented('cubBayesNet_g: kernel order not yet supported')

        return kernFunc



import numpy as np
from .abstract_integrand import AbstractIntegrand
from ..true_measure import Uniform
from ..discrete_distribution import DigitalNetB2


class FourBranch2d(AbstractIntegrand):
    r"""
    Four Branch function in $d=2$. 

    $$g(\boldsymbol{t}) = \min \begin{cases} 3+0.1(t_0-t_1)^2-\frac{t_0-t_1}{\sqrt{2}} \\ 3+0.1(t_0-t_1)^2+\frac{t_0-t_1}{\sqrt{2}} \\ t_0-t_1 + 7/\sqrt{2} \\ t_1-t_0 + 7/\sqrt{2}\end{cases}, \qquad \boldsymbol{T}=(T_0,T_1) \sim \mathcal{U}[-8,8]^2.$$
    
    Examples:
        >>> integrand = FourBranch2d(DigitalNetB2(2,seed=7))
        >>> y = integrand(2**10)
        >>> print("%.4f"%y.mean())
        -2.4995
        >>> integrand.true_measure
        Uniform (AbstractTrueMeasure)
            lower_bound     -8
            upper_bound     2^(3)
            
        With independent replications

        >>> integrand = FourBranch2d(DigitalNetB2(2,seed=7,replications=2**4))
        >>> y = integrand(2**6)
        >>> y.shape
        (16, 64)
        >>> muhats = y.mean(-1) 
        >>> muhats.shape 
        (16,)
        >>> print("%.4f"%muhats.mean())
        -2.5042
    """

    def __init__(self, sampler):
        r"""
        Args:
            sampler (Union[AbstractDiscreteDistribution,AbstractTrueMeasure]): Either  
                
                - a discrete distribution from which to transform samples, or
                - a true measure by which to compose a transform.
        """
        self.sampler = sampler
        assert self.sampler.d==2
        self.true_measure = Uniform(self.sampler,lower_bound=-8,upper_bound=8)
        super(FourBranch2d,self).__init__(dimension_indv=(),dimension_comb=(),parallel=False)
    
    def g(self, t):
        t0,t1 = t[...,0],t[...,1]
        return np.minimum.reduce([
            3+.1*(t0-t1)**2-(t0+t1)/np.sqrt(2),
            3+.1*(t0-t1)**2+(t0+t1)/np.sqrt(2),
            t0-t1+7/np.sqrt(2),
            t1-t0+7/np.sqrt(2)])
    
    def _spawn(self, level, sampler):
        return FourBranch2d(sampler=sampler) 



from .abstract_integrand import AbstractIntegrand
from ..discrete_distribution import DigitalNetB2
from ..true_measure import Gaussian, Uniform
import numpy as np

class CustomFun(AbstractIntegrand):
    r"""
    User supplied integrand $g$. In the following example we implement 

    Examples:
        First we will implement 

        $$g(\boldsymbol{t}) = t_1^2t_2, \qquad \boldsymbol{T}=(T_1,T_2) \sim \mathcal{N}((1,2)^T,\mathsf{I}).$$

        >>> integrand = CustomFun(
        ...     true_measure = Gaussian(DigitalNetB2(2,seed=7),mean=[1,2]),
        ...     g = lambda t: t[...,0]**2*t[...,1])
        >>> y = integrand(2**10)
        >>> print("%.4f"%y.mean())
        3.9991

        With independent replications

        >>> integrand = CustomFun(
        ...     true_measure = Gaussian(DigitalNetB2(2,seed=7,replications=2**4),mean=[1,2]),
        ...     g = lambda t: t[...,0]**2*t[...,1])
        >>> y = integrand(2**6)
        >>> y.shape
        (16, 64)
        >>> muhats = y.mean(1) 
        >>> muhats.shape 
        (16,)
        >>> print("%.4f"%muhats.mean())
        3.9330

        Next we will implement the multi-output function 

        $$g(\boldsymbol{t}) = \begin{pmatrix} \sin(t_1)\cos(t_2) \\ \cos(t_1)\sin(t_2) \\ \sin(t_1)+\cos(t_2) \\ \cos(t_1)+\sin(t_2) \end{pmatrix} \qquad \boldsymbol{T}=(T_1,T_2) \sim \mathcal{U}[0,2\pi]^2.$$
        
        >>> def g(t):
        ...     t1,t2 = t[...,0],t[...,1]
        ...     sint1,cost1,sint2,cost2 = np.sin(t1),np.cos(t1),np.sin(t2),np.cos(t2)
        ...     y1 = sint1*cost2
        ...     y2 = cost1*sint2
        ...     y3 = sint1+cost2
        ...     y4 = cost1+sint2
        ...     y = np.stack([y1,y2,y3,y4])
        ...     return y
        >>> integrand = CustomFun(
        ...     true_measure = Uniform(DigitalNetB2(2,seed=7),lower_bound=0,upper_bound=2*np.pi),
        ...     g = g, 
        ...     dimension_indv = (4,))
        >>> x = integrand.discrete_distrib(2**10)
        >>> y = integrand.f(x)
        >>> y.shape
        (4, 1024)
        >>> with np.printoptions(formatter={"float": lambda x: "%.2e"%x}):
        ...     y.mean(-1)
        array([8.18e-04, 1.92e-06, -2.26e-10, 5.05e-07])

        Stopping criterion which supporting vectorized outputs may pass in Boolean `compute_flags` with `dimension_indv` shape indicating which output need to evaluated, 
            i.e. where `compute_flags` is `False` we do not need to evaluate the integrand. We have not used this in inexpensive example above. 
    
        With independent replications

        >>> integrand = CustomFun(
        ...     true_measure = Uniform(DigitalNetB2(2,seed=7,replications=2**4),lower_bound=0,upper_bound=2*np.pi),
        ...     g = g, 
        ...     dimension_indv = (4,))
        >>> x = integrand.discrete_distrib(2**6)
        >>> x.shape
        (16, 64, 2)
        >>> y = integrand.f(x)
        >>> y.shape
        (4, 16, 64)
        >>> muhats = y.mean(-1) 
        >>> muhats.shape 
        (4, 16)
        >>> with np.printoptions(formatter={"float": lambda x: "%.2e"%x}):
        ...     muhats.mean(-1)
        array([3.83e-03, -6.78e-03, -1.56e-03, -5.65e-04])

    """

    def __init__(self, true_measure, g, dimension_indv=(), parallel=False):
        """
        Args:
            true_measure (AbstractTrueMeasure): The true measure. 
            g (callable): A function handle. 
            dimension_indv (tuple): Shape of individual solution outputs from `g`.
            parallel (int): Parallelization flag. 
                
                - When `parallel = 0` or `parallel = 1` then function evaluation is done in serial fashion.
                - `parallel > 1` specifies the number of processes used by `multiprocessing.Pool` or `multiprocessing.pool.ThreadPool`.
            
                Setting `parallel=True` is equivalent to `parallel = os.cpu_count()`.
        
        Note:
            For `parallel > 1` do *not* set `g` to be anonymous function (i.e. a `lambda` function)
        """
        self.parameters = []
        self.true_measure = true_measure
        self.sampler = self.true_measure
        self.__g = g            
        super(CustomFun,self).__init__(dimension_indv=dimension_indv,dimension_comb=dimension_indv,parallel=parallel)
    
    def g(self, t, *args, **kwargs):
        return self.__g(t,*args,**kwargs)
    
    def _spawn(self, level, sampler):
        return CustomFun(
            true_measure = sampler,
            g = self.__g,
            dimension_indv = self.d_indv,
            parallel = self.parallel)



from ..util import ParameterError, ParameterWarning, MethodImplementationError, _univ_repr, DimensionError
import warnings
import numpy as np
from typing import Union 

class AbstractDiscreteDistribution(object):

    def __init__(self, dimension, replications, seed, d_limit, n_limit):
        self.mimics = 'StdUniform'
        if not hasattr(self,'parameters'):
            self.parameters = []
        self.d_limit = d_limit
        self.n_limit = n_limit 
        if not (np.isscalar(self.d_limit) and self.d_limit>0 and np.isscalar(self.n_limit) and self.n_limit>0):
            raise ParameterError("d_limit and n_limit must be greater than 0")
        self.d_limit = np.inf if self.d_limit==np.inf else int(self.d_limit)
        self.n_limit = np.inf if self.n_limit==np.inf else int(self.n_limit)
        self.no_replications = replications is None 
        self.replications = 1 if self.no_replications else int(replications)
        if self.replications<0:
            raise ParameterError("replications must be None or a postive int")
        if isinstance(dimension,list) or isinstance(dimension,tuple) or isinstance(dimension,np.ndarray):
            self.dvec = np.array(dimension,dtype=int)
            self.d = len(self.dvec)
            if not (self.dvec.ndim==1 and len(np.unique(self.dvec))==self.d):
                raise ParameterError("dimension must be a 1d array of unique values")
        else:
            self.d = int(dimension)
            self.dvec = np.arange(self.d)
        if any(self.dvec>self.d_limit):
            raise ParameterError('dimension greater than dimension limit %d'%self.d_limit)
        self._base_seed = seed if isinstance(seed,np.random.SeedSequence) else np.random.SeedSequence(seed)
        self.entropy = self._base_seed.entropy
        self.spawn_key = self._base_seed.spawn_key
        self.rng = np.random.Generator(np.random.SFC64(self._base_seed))

    def __call__(self, n=None, n_min=None, n_max=None, return_binary=False, warn=True):
        r"""
        - If just `n` is supplied, generate samples from the sequence at indices 0,...,`n`-1.
        - If `n_min` and `n_max` are supplied, generate samples from the sequence at indices `n_min`,...,`n_max`-1.
        - If `n` and `n_min` are supplied, then generate samples from the sequence at indices `n`,...,`n_min`-1.

        Args:
            n (Union[None,int]): Number of points to generate.
            n_min (Union[None,int]): Starting index of sequence.
            n_max (Union[None,int]): Final index of sequence.
            return_binary (bool): Only used for `DigitalNetB2`.  
                If `True`, *only* return the integer representation `x_integer` of base 2 digital net.  
            warn (bool): If `False`, disable warnings when generating samples.

        Returns:
            x (np.ndarray): Samples from the sequence. 
                
                - If `replications` is `None` then this will be of size (`n_max`-`n_min`) $\times$ `dimension` 
                - If `replications` is a positive int, then `x` will be of size `replications` $\times$ (`n_max`-`n_min`) $\times$ `dimension` 

                Note that if `return_binary=True` then `x` is returned where `x` are integer representations of the digital net points. 
        """
        return self.gen_samples(n=n,n_min=n_min,n_max=n_max,return_binary=return_binary,warn=warn)

    def gen_samples(self, n=None, n_min=None, n_max=None, return_binary=False, warn=True):
        if n is not None and n_min is None and n_max is None: 
            n_min = 0 
            n_max = int(n)
        elif n is None and n_min is not None and n_max is not None: 
            n_min = int(n_min) 
            n_max = int(n_max)
        elif n is not None and n_min is not None and n_max is None:
            n_max = int(n_min) 
            n_min = int(n)
        else:
            raise ParameterError("Please provide either n or (n_min,n_max)")
        if not (0<=n_min and n_min<=n_max and n_max<=self.n_limit):
            raise ParameterError("require 0 <= n_min (%d) <= n_max (%d) <= n_limit (%d)"%(n_min,n_max,self.n_limit))
        x = self._gen_samples(n_min=n_min,n_max=n_max,return_binary=return_binary,warn=warn)
        if isinstance(x,np.ndarray):
            if not x.shape==(self.replications,n_max-n_min,self.d):
                raise ParameterError("x returned by _gen_samples must have shape self.replications (%d) x n (%d) x d (%d) but got shape %s"%(self.replications,n_max-n_min,self.d,str(x.shape)))
            return x[0] if self.no_replications else x
        if not isinstance(x,tuple):
            raise ParameterError("returned value from _gen_samples must be a np.ndarray or tuple of them")
        for xi in x:
            if not xi.shape==(self.replications,n_max-n_min,self.d):
                raise ParameterError("xs returned by _gen_samples must each have shape self.replications (%d) x n (%d) x d (%d) but got shape %s"%(self.replications,n_max-n_min,self.d,str(x.shape)))
        return (xi[0] if self.no_replications else xi for xi in x)
    
    def _gen_samples(self, *args, **kwargs):
        raise MethodImplementationError(self, '_gen_samples')
    
    def spawn(self, s=1, dimensions=None):
        r"""
        Spawn new instances of the current discrete distribution but with new seeds and dimensions.
        Used by multi-level QMC algorithms which require different seeds and dimensions on each level.

        Note:
            Use `replications` instead of using `spawn` when possible, e.g., when spawning copies which all have the same dimension.

        Args:
            s (int): Number of copies to spawn
            dimensions (np.ndarray): Length `s` array of dimensions for each copy. Defaults to the current dimension. 
        
        Returns:
            spawned_discrete_distribs (list): Discrete distributions with new seeds and dimensions.
        """
        s = int(s)
        if s<=0:
            raise ParameterError("Must spawn s>0 instances")
        if dimensions is None: 
            dimensions = np.tile(self.d,s)
        elif (isinstance(dimensions,list) or isinstance(dimensions,tuple) or isinstance(dimensions,np.ndarray)):
            dimensions = np.array(dimensions,dtype=int)
        else:
            dimensions = np.tile(dimensions,s)
        if not (dimensions.ndim==1 and len(dimensions)==s):
            raise ParameterError("dimensions must be a length s np.ndarray")
        child_seeds = self._base_seed.spawn(s)
        spawned_discrete_distribs = [self._spawn(child_seeds[i],int(dimensions[i])) for i in range(s)]
        return spawned_discrete_distribs

    def _spawn(self, child_seed, dimension):
        raise MethodImplementationError(self, '_spawn')

    def pdf(self, x):
        return np.ones_like(x[...,0])
    
    def __repr__(self, abc_class_name):
        return _univ_repr(self, abc_class_name, ['d','replications']+self.parameters+['entropy'])


class AbstractLDDiscreteDistribution(AbstractDiscreteDistribution):
    """ Low discrepancy sequence. Alias for `AbstractDiscreteDistribution` used for compatibility checks."""
    
    def __repr__(self):
        return super().__repr__("AbstractLDDiscreteDistribution")


class AbstractIIDDiscreteDistribution(AbstractDiscreteDistribution):
    """ IID sequence. Alias for `AbstractDiscreteDistribution` used for compatibility checks."""
    
    def __repr__(self):
        return super().__repr__("AbstractIIDDiscreteDistribution")



from .abstract_stopping_criterion import AbstractStoppingCriterion
from ..util.data import Data

from ..util import MaxSamplesWarning, ParameterError, ParameterWarning, CubatureWarning
from ..integrand import AbstractIntegrand
import numpy as np
from time import time
import warnings


class AbstractCubQMCLDG(AbstractStoppingCriterion):
    
    def __init__(self, integrand, abs_tol, rel_tol, n_init, n_limit, fudge, check_cone,
        control_variates, control_variate_means, update_beta, ptransform,
        ft, omega, allowed_distribs, cast_complex, error_fun):
        self.parameters = ['abs_tol','rel_tol','n_init','n_limit']
        # Input Checks
        if np.log2(n_init)%1!=0 or n_init<2**8:
            warnings.warn('n_init must be a power of two at least 2**8. Using n_init = 2**8',ParameterWarning)
            n_init = 2**8
        if np.log2(n_limit)%1!=0:
            warnings.warn('n_init must be a power of two. Using n_limit = 2**30',ParameterWarning)
            n_limit = 2**30
        # Set Attributes
        self.n_init = int(n_init)
        self.m_init = int(np.log2(n_init))
        self.n_limit = int(n_limit)
        assert isinstance(error_fun,str) or callable(error_fun)
        if isinstance(error_fun,str):
            if error_fun.upper()=="EITHER":
                error_fun = lambda sv,abs_tol,rel_tol: np.maximum(abs_tol,abs(sv)*rel_tol)
            elif error_fun.upper()=="BOTH":
                error_fun = lambda sv,abs_tol,rel_tol: np.minimum(abs_tol,abs(sv)*rel_tol)
            else:
                raise ParameterError("str error_fun must be 'EITHER' or 'BOTH'")
        self.error_fun = error_fun
        self.fudge = fudge
        self.check_cone = check_cone
        self.ft = ft
        self.omega = omega
        self.ptransform = ptransform
        self.cast_complex = cast_complex
        self.r_lag = 4
        self.omg_circ = lambda m: 2**(-m)
        self.l_star = int(self.m_init-self.r_lag)
        self.omg_hat = lambda m: self.fudge(m)/((1+self.fudge(self.r_lag))*self.omg_circ(self.r_lag))
        # QMCPy Objs
        self.integrand = integrand
        self.true_measure = self.integrand.true_measure
        self.discrete_distrib = self.integrand.discrete_distrib
        super(AbstractCubQMCLDG,self).__init__(allowed_distribs=allowed_distribs,allow_vectorized_integrals=True)
        assert self.integrand.discrete_distrib.no_replications==True, "Require the discrete distribution has replications=None"
        assert self.integrand.discrete_distrib.randomize!="FALSE", "Require discrete distribution is randomized"
        self.set_tolerance(abs_tol,rel_tol)
        # control variates
        self.cv_mu = np.atleast_1d(control_variate_means)
        self.cv = control_variates
        if isinstance(self.cv,AbstractIntegrand):
            self.cv = [self.cv]
            self.cv_mu = self.cv_mu[None,...]
        assert isinstance(self.cv,list), "cv must be a list of AbstractIntegrand objects"
        for cv in self.cv:
            if (not isinstance(cv,AbstractIntegrand)) or (cv.discrete_distrib!=self.discrete_distrib) or (cv.d_indv!=self.integrand.d_indv):
                raise ParameterError('''
                        Each control variates discrete distribution must be an AbstractIntegrand instance 
                        with the same discrete distribution as the main integrand. d_indv must also match 
                        that of the main integrand instance for each control variate.''')
        self.ncv = len(self.cv)
        self.update_beta = update_beta
        if self.ncv>0:
            assert self.cv_mu.shape==((self.ncv,)+self.integrand.d_indv), "Control variate means should have shape (len(control variates),d_indv)."
            self.parameters += ['cv','cv_mu','update_beta']
        else:
            self.update_beta = False
        self.vlstsq = np.vectorize(lambda x,y: np.linalg.lstsq(x.T,y,rcond=None)[0],signature="(k,m),(m)->(k)")
    
    def _update_kappanumap(self, kappanumap, ytildefull, mfrom, mto, m):
        for l in range(int(mfrom),int(mto),-1):
            nl = 2**l
            oldone = np.abs(np.take_along_axis(ytildefull,kappanumap[...,1:int(nl)],axis=-1)) # earlier values of kappa, don't touch first one
            newone = np.abs(np.take_along_axis(ytildefull,kappanumap[...,nl+1:2*nl],axis=-1)) # later values of kappa,
            *prioridxs,flip = np.where(newone>oldone)
            flip = flip+1 # add one to account for the fact that we do not consider indices which are powers of 2
            if flip.size!=0:
                additive = np.arange(0,2**m-1,2**(l+1))
                flipall = (flip[None,:]+additive[:,None]).flatten()
                zeroadditive = np.zeros(len(additive),dtype=int) 
                pidxs = tuple((pidx[None,:]+zeroadditive[:,None]).flatten() for pidx in prioridxs) # alternative to tiling
                kappanumap[pidxs+(flipall,)],kappanumap[pidxs+(nl+flipall,)] = kappanumap[pidxs+(nl+flipall,)],kappanumap[pidxs+(flipall,)]
        return kappanumap
    
    def _beta_update(self, beta, kappanumap, ytildefull, ycvtildefull, mstart):
        kappa_approx = kappanumap[...,(2**mstart):] # kappa index used for fitting
        y4beta = np.take_along_axis(ytildefull,kappa_approx,axis=-1)
        x4beta = np.take_along_axis(ycvtildefull,kappa_approx[...,None,:],axis=-1)
        beta = self.vlstsq(x4beta,y4beta)
        return beta
    
    def integrate(self):
        t_start = time()
        data = Data(
            parameters = [
                'solution',
                'comb_bound_low',
                'comb_bound_high',
                'comb_bound_diff',
                'comb_flags',
                'n_total',
                'n',
                'time_integrate'])
        data.flags_indv = np.tile(False,self.integrand.d_indv)
        data.compute_flags = np.tile(True,self.integrand.d_indv)
        data.n = np.tile(self.n_init,self.integrand.d_indv)
        data.n_min = 0 
        data.n_max = self.n_init
        data.solution_indv = np.tile(np.nan,self.integrand.d_indv)
        data.xfull = np.empty((0,self.integrand.d))
        data.yfull = np.empty(self.integrand.d_indv+(0,))
        if self.ncv>0:
            data.ycvfull = np.empty(self.integrand.d_indv+(self.ncv,0))
        data.bounds_half_width = np.tile(np.inf,self.integrand.d_indv)
        data.muhat = np.tile(np.nan,self.integrand.d_indv)
        data.beta = np.tile(np.nan,self.integrand.d_indv+(self.ncv,))
        while True:
            m = int(np.log2(data.n_max))
            xnext = self.discrete_distrib(n_min=data.n_min,n_max=data.n_max)
            data.xfull = np.concatenate([data.xfull,xnext],0)
            ynext = self.integrand.f(xnext,periodization_transform=self.ptransform,compute_flags=data.compute_flags)
            ynext[~data.compute_flags] = np.nan
            data.yfull = np.concatenate([data.yfull,ynext],-1)
            if self.ncv>0:
                ycvnext = [None]*self.ncv
                for k in range(self.ncv):
                    ycvnext_k = self.cv[k].f(xnext,periodization_transform=self.ptransform,compute_flags=data.compute_flags)
                    ycvnext_k[~data.compute_flags] = np.nan
                    ycvnext[k] = ycvnext_k
                ycvnext = np.stack(ycvnext,-2)
                data.ycvfull = np.concatenate([data.ycvfull,ycvnext],-1)
            mllstart = m-self.r_lag-1
            nllstart = 2**mllstart
            if data.n_min==0: # first iteration
                n = int(2**m)
                ytildefull = self.ft(ynext)/np.sqrt(n)
                kappanumap = self._update_kappanumap(np.tile(np.arange(n),self.integrand.d_indv+(1,)),ytildefull,m-1,0,m)
                if self.ncv>0:
                    ycvtildefull = self.ft(ycvnext)/np.sqrt(n)
                    data.beta = self._beta_update(data.beta,kappanumap,ytildefull,ycvtildefull,mllstart)
                    ytildefull = ytildefull-(ycvtildefull*data.beta[...,None]).sum(-2)
                    kappanumap = self._update_kappanumap(np.tile(np.arange(n),self.integrand.d_indv+(1,)),ytildefull,m-1,0,m)
            else: # any iteration after the first
                mnext = int(m-1)
                n = int(2**mnext)
                if not self.update_beta: # do not update the beta coefficients
                    if self.ncv>0:
                        ynext[data.compute_flags] = ynext[data.compute_flags]-(ycvnext[data.compute_flags]*data.beta[data.compute_flags,:,None]).sum(-2)
                    ytildeomega = self.omega(mnext)*self.ft(ynext[data.compute_flags])/np.sqrt(n)
                    ytildefull_next = np.nan*np.ones_like(ytildefull)
                    ytildefull_next[data.compute_flags] = (ytildefull[data.compute_flags]-ytildeomega)/2
                    ytildefull[data.compute_flags] = (ytildefull[data.compute_flags]+ytildeomega)/2
                    ytildefull = np.concatenate([ytildefull,ytildefull_next],axis=-1)
                else: # update beta
                    ytildefull = np.concatenate([ytildefull,np.tile(np.nan,ytildefull.shape)],axis=-1)
                    ytildefull[data.compute_flags] = self.ft(data.yfull[data.compute_flags])/np.sqrt(2**m)
                    ycvtildefull = np.concatenate([ycvtildefull,np.tile(np.nan,ycvtildefull.shape)],axis=-1)
                    ycvtildefull[data.compute_flags] = self.ft(data.ycvfull[data.compute_flags])/np.sqrt(2**m)
                    data.beta[data.compute_flags] = self._beta_update(data.beta[data.compute_flags],kappanumap[data.compute_flags],ytildefull[data.compute_flags],ycvtildefull[data.compute_flags],mllstart)
                kappanumap = np.concatenate([kappanumap,n+kappanumap],axis=-1)
                kappanumap[data.compute_flags] = self._update_kappanumap(kappanumap[data.compute_flags],ytildefull[data.compute_flags],m-1,mllstart,m)
            if self.ncv==0:
                data.muhat[data.compute_flags] = data.yfull[data.compute_flags].mean(-1)
            else:
                ydiff = data.yfull[data.compute_flags]-(data.ycvfull[data.compute_flags]*data.beta[data.compute_flags,:,None]).sum(-2)
                data.muhat[data.compute_flags] = ydiff.mean(-1)+(data.beta[data.compute_flags]*np.moveaxis(self.cv_mu,0,-1)[data.compute_flags]).sum(-1)
            data.bounds_half_width[data.compute_flags] = self.fudge(m)*np.abs(np.take_along_axis(ytildefull[data.compute_flags],kappanumap[data.compute_flags][...,nllstart:2*nllstart],axis=-1)).sum(-1)
            data.indv_bound_low = data.muhat-data.bounds_half_width
            data.indv_bound_high = data.muhat+data.bounds_half_width
            if self.check_cone:
                data.c_stilde_low = np.tile(-np.inf,(m+1-self.l_star,)+self.integrand.d_indv)
                data.c_stilde_up = np.tile(np.inf,(m+1-self.l_star,)+self.integrand.d_indv)
                for l in range(self.l_star,m+1): # Storing the information for the necessary conditions
                    c_tmp = self.omg_hat(m-l)*self.omg_circ(m-l)
                    c_low = 1./(1+c_tmp)
                    c_up = 1./(1-c_tmp)
                    const1 = np.abs(np.take_along_axis(ytildefull[data.compute_flags],kappanumap[data.compute_flags][...,int(2**(l-1)):int(2**l)],axis=-1)).sum(-1)
                    idx = int(l-self.l_star)
                    data.c_stilde_low[idx,data.compute_flags] = np.maximum(data.c_stilde_low[idx,data.compute_flags],c_low*const1)
                    if c_tmp < 1:
                        data.c_stilde_up[idx,data.compute_flags] = np.minimum(data.c_stilde_up[idx,data.compute_flags],c_up*const1)
                data.cone_violation = (data.c_stilde_low > data.c_stilde_up).any(0)
                if data.cone_violation.sum()>0:
                    warnings.warn('Cone condition violations at indices in data.cone_violation.',CubatureWarning)
            else:
                data.cone_violation = None
            data.n[data.compute_flags] = data.n_max
            data.n_total = data.n_max
            data.comb_bound_low,data.comb_bound_high = self.integrand.bound_fun(data.indv_bound_low,data.indv_bound_high)
            data.comb_bound_diff = data.comb_bound_high-data.comb_bound_low
            fidxs = np.isfinite(data.comb_bound_low)&np.isfinite(data.comb_bound_high)
            slow,shigh,abs_tols,rel_tols = data.comb_bound_low[fidxs],data.comb_bound_high[fidxs],self.abs_tols[fidxs],self.rel_tols[fidxs]
            data.solution = np.tile(np.nan,data.comb_bound_low.shape)
            data.solution[fidxs] = 1/2*(slow+shigh+self.error_fun(slow,abs_tols,rel_tols)-self.error_fun(shigh,abs_tols,rel_tols))
            data.comb_flags = np.tile(False,data.comb_bound_low.shape)
            data.comb_flags[fidxs] = (shigh-slow) <= (self.error_fun(slow,abs_tols,rel_tols)+self.error_fun(shigh,abs_tols,rel_tols))
            data.flags_indv = self.integrand.dependency(data.comb_flags)
            data.compute_flags = ~data.flags_indv
            if np.sum(data.compute_flags)==0:
                break # sufficiently estimated
            elif 2*data.n_total>self.n_limit:
                warning_s = """
                Already generated %d samples.
                Trying to generate %d new samples would exceeds n_limit = %d.
                No more samples will be generated.
                Note that error tolerances may not be satisfied. """ \
                % (int(data.n_total),int(data.n_total),int(self.n_limit))
                warnings.warn(warning_s, MaxSamplesWarning)
                break
            data.n_min = data.n_max
            data.n_max = 2*data.n_min
        data.stopping_crit = self
        data.integrand = self.integrand
        data.true_measure = self.integrand.true_measure
        data.discrete_distrib = self.true_measure.discrete_distrib
        data.time_integrate = time()-t_start
        return data.solution,data
    
    def set_tolerance(self, abs_tol=None, rel_tol=None, rmse_tol=None):
        assert rmse_tol is None, "rmse_tol not supported by this stopping criterion."
        if abs_tol is not None:
            self.abs_tol = abs_tol
            self.abs_tols = np.full(self.integrand.d_comb,self.abs_tol)
        if rel_tol is not None:
            self.rel_tol = rel_tol
            self.rel_tols = np.full(self.integrand.d_comb,self.rel_tol)



import numpy as np 


class Polynomial():
    """
    >>> p = Polynomial([3,5,0,1,2])
    >>> x = np.random.rand(3,4)
    >>> y = p(x) 
    >>> y_true = 3*x**4 + 5*x**3 + 0*x**2 + 1*x**1 + 2*x**0
    >>> assert np.allclose(y,y_true,atol=1e-12)
    """
    def __init__(self, coeffs):
        """
        Polynomial evaluation with Horner's rule
        
        Args:
            coeffs (list or np.ndarray or torch.Tensor): vector of coefficients  
            e.g. coeffs = [a, b, c] corresponds to the quadratic polynomial a*x**2 + b*x + c
        """
        assert isinstance(coeffs,list)
        self.order = len(coeffs) 
        assert self.order >= 1
        self.coeffs = coeffs
    def __call__(self, x):
        if isinstance(x,np.ndarray):
            npt = np 
            powers = np.arange(self.order-1,-1,-1,dtype=x.dtype)
            coeffs = np.array(self.coeffs,dtype=x.dtype)
        else:
            import torch 
            npt = torch
            powers = torch.arange(self.order-1,-1,-1,dtype=x.dtype,device=x.device)
            coeffs = torch.tensor(self.coeffs,dtype=x.dtype,device=x.device)
        return (x[...,None]**powers*coeffs).sum(-1)

BERNOULLIPOLYSDICT = {
    #0:  Polynomial([1]),
    1:  Polynomial([1, -1/2]),
    2:  Polynomial([1, -1,   1/6]),
    3:  Polynomial([1, -3/2, 1/2,  0]),
    4:  Polynomial([1, -2,   1,    0, -1/30]),
    5:  Polynomial([1, -5/2, 5/3,  0, -1/6,  0]),
    6:  Polynomial([1, -3,   5/2,  0, -1/2,  0, 1/42]),
    7:  Polynomial([1, -7/2, 7/2,  0, -7/6,  0, 1/6, 0]),
    8:  Polynomial([1, -4,   14/3, 0, -7/3,  0, 2/3, 0, -1/30]),
    9:  Polynomial([1, -9/2, 6,    0, -21/5, 0, 2,   0, -3/10, 0]),
    10: Polynomial([1, -5,   15/2, 0, -7,    0, 5,   0, -3/2,  0, 5/66])
}

def bernoulli_poly(n, x):
    r"""
    $n^\text{th}$ Bernoulli polynomial

    Examples:
        >>> x = np.arange(6).reshape((2,3))/6
        >>> available_n = list(BERNOULLIPOLYSDICT.keys())
        >>> available_n
        [1, 2, 3, 4, 5, 6, 7, 8, 9, 10]
        >>> for n in available_n:
        ...     y = bernoulli_poly(n,x)
        ...     with np.printoptions(precision=2):
        ...         print("n = %d\n%s"%(n,y))
        n = 1
        [[-0.5  -0.33 -0.17]
         [ 0.    0.17  0.33]]
        n = 2
        [[ 0.17  0.03 -0.06]
         [-0.08 -0.06  0.03]]
        n = 3
        [[ 0.    0.05  0.04]
         [ 0.   -0.04 -0.05]]
        n = 4
        [[-0.03 -0.01  0.02]
         [ 0.03  0.02 -0.01]]
        n = 5
        [[ 0.00e+00 -2.19e-02 -2.06e-02]
         [ 1.39e-17  2.06e-02  2.19e-02]]
        n = 6
        [[ 0.02  0.01 -0.01]
         [-0.02 -0.01  0.01]]
        n = 7
        [[ 0.00e+00  2.28e-02  2.24e-02]
         [-1.39e-17 -2.24e-02 -2.28e-02]]
        n = 8
        [[-0.03 -0.02  0.02]
         [ 0.03  0.02 -0.02]]
        n = 9
        [[ 0.   -0.04 -0.04]
         [ 0.    0.04  0.04]]
        n = 10
        [[ 0.08  0.04 -0.04]
         [-0.08 -0.04  0.04]]
        >>> import scipy.special
        >>> for n in available_n:
        ...     bpoly_coeffs = BERNOULLIPOLYSDICT[n].coeffs
        ...     bpoly_coeffs_true = scipy.special.bernoulli(n)*scipy.special.comb(n,np.arange(n,-1,-1))
        ...     assert np.allclose(bpoly_coeffs_true,bpoly_coeffs,atol=1e-12)
    
    Args:
        n (int): Polynomial order.
        x (Union[np.ndarray,torch.Tensor]): Points at which to evaluate the Bernoulli polynomial.
    
    Returns:
        y (Union[np.ndarray,torch.Tensor]): Bernoulli polynomial values.
    """
    assert isinstance(n,int)
    assert n in BERNOULLIPOLYSDICT, "n = %d not in BERNOULLIPOLYSDICT"%n
    bpoly = BERNOULLIPOLYSDICT[n]
    y = bpoly(x) 
    return y


import numpy as np
from .abstract_integrand import AbstractIntegrand
from ..true_measure import Uniform
from ..discrete_distribution import DigitalNetB2

class Multimodal2d(AbstractIntegrand):
    r"""
    Multimodal function in $d=2$ dimensions. 

    $$g(\boldsymbol{t}) = (t_0^2+4)(t_1-1)/20-\sin(5t_0/2)-2 \qquad \boldsymbol{T} = (T_0,T_1) \sim \mathcal{U}([-4,7] \times [-3,8]).$$

    Examples:
        >>> integrand = Multimodal2d(DigitalNetB2(2,seed=7))
        >>> y = integrand(2**10)
        >>> print("%.4f"%y.mean())
        -0.7365
        >>> integrand.true_measure
        Uniform (AbstractTrueMeasure)
            lower_bound     [-4 -3]
            upper_bound     [7 8]

        With independent replications

        >>> integrand = Multimodal2d(DigitalNetB2(2,seed=7,replications=2**4))
        >>> y = integrand(2**6)
        >>> y.shape
        (16, 64)
        >>> muhats = y.mean(-1) 
        >>> muhats.shape 
        (16,)
        >>> print("%.4f"%muhats.mean())
        -0.7366
    """
    def __init__(self, sampler):
        r"""
        Args:
            sampler (Union[AbstractDiscreteDistribution,AbstractTrueMeasure]): Either  
                
                - a discrete distribution from which to transform samples, or
                - a true measure by which to compose a transform.
        """
        self.sampler = sampler 
        assert self.sampler.d==2
        self.true_measure = Uniform(self.sampler,lower_bound=[-4,-3],upper_bound=[7,8])
        super(Multimodal2d,self).__init__(dimension_indv=(),dimension_comb=(),parallel=False)
    def g(self, t):
        t0,t1 = t[...,0],t[...,1]
        return (t0**2+4)*(t1-1)/20-np.sin(5*t0/2)-2
    def _spawn(self, level, sampler):
        return Multimodal2d(sampler=sampler) 



from ..util import MethodImplementationError
import numpy as np 
from typing import Union
from ..util.transforms import tf_exp_eps,tf_exp_eps_inv,parse_assign_param,tf_identity,insert_batch_dims


class AbstractKernel(object):

    def __new__(cls, *args, **kwargs):
        if (
            ("torchify" in kwargs and kwargs["torchify"]) or 
            ("base_kernel" in kwargs and kwargs["base_kernel"].torchify) or 
            (len(args)>0 and isinstance(args[0],AbstractKernel) and args[0].torchify)
            ):
            import torch 
            x = type(cls.__name__,(cls,torch.nn.Module),{})
            instance = super().__new__(x)
        else:
            instance = super().__new__(cls)
        return instance

    def __init__(self, d, torchify, device, compile_call, comiple_call_kwargs):
        super().__init__()
        # dimension 
        assert d%1==0 and d>0, "dimension d must be a positive int"
        self.d = d
        # torchify
        self.torchify = torchify 
        if self.torchify: 
            import torch 
            self.npt = torch
            self.nptarray = torch.tensor
            self.nptarraytype = torch.Tensor
            self.device = torch.device(device) 
            self.nptkwargs = {"device":device}
        else:
            self.npt = np
            self.nptarray = np.array
            self.nptarraytype = np.ndarray
            self.device = None
            self.nptkwargs = {}
        self.batch_param_names = []
        if compile_call:
            assert self.torchify, "compile_call requires torchify is True"
            import torch
            self.compiled_parsed___call__ = torch.compile(self.parsed___call__,**comiple_call_kwargs)
        else:
            self.compiled_parsed___call__ = self.parsed___call__
    
    @property 
    def nbdim(self):
        empty = self.npt.empty((0,self.d),**self.nptkwargs)
        v = self.__call__(empty,empty)
        nbdim = v.ndim-1
        return nbdim
    
    @property 
    def batch_params(self):
        return {pname:getattr(self,pname) for pname in self.batch_param_names}
    
    def get_batch_params(self, ndim):
        return {pname: insert_batch_dims(batch_param,ndim,-1) for pname,batch_param in self.batch_params.items()}
    
    def __call__(self, x0, x1, beta0=None, beta1=None, c=None):
        r"""
        Evaluate the kernel with (optional) partial derivatives 

        $$\sum_{\ell=1}^p c_{\ell} \partial_{\boldsymbol{x}_0}^{\boldsymbol{\beta}_{\ell 0}} \partial_{\boldsymbol{x}_1}^{\boldsymbol{\beta}_{\ell 1}} K(\boldsymbol{x}_0,\boldsymbol{x}_1).$$
        
        Args:
            x0 (Union[np.ndarray,torch.Tensor]): Shape `x0.shape=(...,d)` first input to kernel with 
            x1 (Union[np.ndarray,torch.Tensor]): Shape `x1.shape=(...,d)` second input to kernel with 
            beta0 (Union[np.ndarray,torch.Tensor]): Shape `beta0.shape=(p,d)` derivative orders with respect to first inputs, $\boldsymbol{\beta}_0$.
            beta1 (Union[np.ndarray,torch.Tensor]): Shape `beta1.shape=(p,d)` derivative orders with respect to first inputs, $\boldsymbol{\beta}_1$.
            c (Union[np.ndarray,torch.Tensor]): Shape `c.shape=(p,)` coefficients of derivatives.
        Returns:
            k (Union[np.ndarray,torch.Tensor]): Shape `y.shape=(x0+x1).shape[:-1]` kernel evaluations. 
        """
        assert isinstance(x0,self.nptarraytype) 
        assert isinstance(x0,self.nptarraytype) 
        assert x0.shape[-1]==self.d, "the size of the last dimension of x0 must equal d=%d, got x0.shape=%s"%(self.d,str(tuple(x0.shape)))
        assert x1.shape[-1]==self.d, "the size of the last dimension of x1 must equal d=%d, got x1.shape=%s"%(self.d,str(tuple(x1.shape)))
        if beta0 is None:
            beta0 = self.npt.zeros((1,self.d),dtype=int,**self.nptkwargs)
        if beta1 is None:
            beta1 = self.npt.zeros((1,self.d),dtype=int,**self.nptkwargs)
        if not isinstance(beta0,self.nptarraytype):
            beta0 = self.nptarray(beta0)
        if not isinstance(beta1,self.nptarraytype):
            beta1 = self.nptarray(beta1)
        beta0 = self.npt.atleast_2d(beta0)
        beta1 = self.npt.atleast_2d(beta1)
        assert beta0.ndim==2 and beta1.ndim==2, "beta0 and beta1 must both be 2 dimensional"
        p = beta0.shape[0]
        assert beta0.shape==(p,self.d), "expected beta0.shape=(%d,%d) but got beta0.shape=%s"%(p,self.d,str(tuple(beta0.shape)))
        assert beta1.shape==(p,self.d), "expected beta1.shape=(%d,%d) but got beta1.shape=%s"%(p,self.d,str(tuple(beta1.shape)))
        assert (beta0%1==0).all() and (beta0>=0).all(), "require int beta0 >= 0"
        assert (beta1%1==0).all() and (beta1>=0).all(), "require int beta1 >= 0"
        if c is None:
            c = self.npt.ones(p,**self.nptkwargs)
        if not isinstance(c,self.nptarraytype):
            c = self.nptarray(c) 
        c = self.npt.atleast_1d(c) 
        assert c.shape==(p,), "expected c.shape=(%d,) but got c.shape=%s"%(p,str(tuple(c.shape)))
        if not self.AUTOGRADKERNEL:
            batch_params = self.get_batch_params(max(x0.ndim-1,x1.ndim-1))
            k = self.compiled_parsed___call__(x0,x1,beta0,beta1,c,batch_params)
        else:
            if (beta0==0).all() and (beta1==0).all():
                batch_params = self.get_batch_params(max(x0.ndim-1,x1.ndim-1))
                k = c.sum()*self.compiled_parsed___call__(x0,x1,batch_params)
            else: # requires autograd, so self.npt=torch
                assert self.torchify, "autograd requires torchify=True"
                import torch
                incoming_grad_enabled = torch.is_grad_enabled()
                torch.set_grad_enabled(True)
                incoming_grad_enabled_params = {pname: param.requires_grad for pname,param in self.named_parameters()}
                if not incoming_grad_enabled:
                    for pname,param in self.named_parameters():
                        param.requires_grad_(False)
                if (beta0>0).any():
                    tileshapex0 = tuple(self.npt.ceil(self.npt.tensor(x1.shape[:-1])/self.npt.tensor(x0.shape[:-1])).to(int))
                    x0gs = [self.npt.tile(x0[...,j].clone().requires_grad_(True),tileshapex0) for j in range(self.d)]
                    [x0gj.requires_grad_(True) for x0gj in x0gs]
                    x0g = self.npt.stack(x0gs,dim=-1)
                else:
                    x0g = x0
                if (beta1>0).any():
                    tileshapex1 = tuple(self.npt.ceil(self.npt.tensor(x0.shape[:-1])/self.npt.tensor(x1.shape[:-1])).to(int))
                    x1gs = [self.npt.tile(x1[...,j].clone().requires_grad_(True),tileshapex1) for j in range(self.d)]
                    [x1gj.requires_grad_(True) for x1gj in x1gs]
                    x1g = self.npt.stack(x1gs,dim=-1)
                else:
                    x1g = x1
                batch_params = self.get_batch_params(max(x0.ndim-1,x1.ndim-1))
                k = 0.
                k_base = self.compiled_parsed___call__(x0g,x1g,batch_params)
                for l in range(p):
                    if (beta0[l]>0).any() or (beta1[l]>0).any():
                        k_part = k_base.clone()
                        for j0 in range(self.d):
                            for _ in range(beta0[l,j0]):
                                k_part = torch.autograd.grad(k_part,x0gs[j0],grad_outputs=torch.ones_like(k_part,requires_grad=True),create_graph=True)[0]
                        for j1 in range(self.d):
                            for _ in range(beta1[l,j1]):
                                k_part = torch.autograd.grad(k_part,x1gs[j1],grad_outputs=torch.ones_like(k_part,requires_grad=True),create_graph=True)[0]
                    else:
                        k_part = k_base 
                    k += c[l]*k_part
                if not incoming_grad_enabled:
                    for pname,param in self.named_parameters():
                        param.requires_grad_(incoming_grad_enabled_params[pname])
                if (not incoming_grad_enabled) or ((not any(incoming_grad_enabled_params.values())) and (not x0.requires_grad) and (not x1.requires_grad)):
                    k = k.detach()
                torch.set_grad_enabled(incoming_grad_enabled)
        return k
    
    def parsed___call__(self, *args, **kwargs):
        raise MethodImplementationError(self, 'parsed___call__')

    def single_integral_01d(self, x):
        r"""
        Evaluate the integral of the kernel over the unit cube

        $$\tilde{K}(\boldsymbol{x}) = \int_{[0,1]^d} K(\boldsymbol{x},\boldsymbol{z}) \; \mathrm{d} \boldsymbol{z}.$$
        
        Args:
            x (Union[np.ndarray,torch.Tensor]): Shape `x0.shape=(...,d)` first input to kernel with 
        
        Returns:
            tildek (Union[np.ndarray,torch.Tensor]): Shape `y.shape=x.shape[:-1]` integral kernel evaluations. 
        """
        if self.npt==np:
            assert isinstance(x,np.ndarray) 
        else: # self.npt==torch
            assert isinstance(x,self.npt.Tensor)
        assert x.shape[-1]==self.d, "the size of the last dimension of x must equal d=%d, got x.shape=%s"%(self.d,str(tuple(x.shape)))
        batch_params = self.get_batch_params(x.ndim-1)
        return self.parsed_single_integral_01d(x,batch_params)
    
    def parsed_single_integral_01d(self, x, batch_params):
        raise MethodImplementationError(self, 'parsed_single_integral_01d')
    
    def double_integral_01d(self):
        r"""
        Evaluate the integral of the kernel over the unit cube

        $$\tilde{K} = \int_{[0,1]^d} \int_{[0,1]^d} K(\boldsymbol{x},\boldsymbol{z}) \; \mathrm{d} \boldsymbol{x} \; \mathrm{d} \boldsymbol{z}.$$
        
        Returns:
            tildek (Union[np.ndarray,torch.Tensor]): Double integral kernel evaluations.
        """
        raise MethodImplementationError(self, 'double_integral_01d')

    def rel_pairwise_dist_func(self, x0, x1, lengthscales):
        return self.npt.linalg.norm((x0-x1)/(np.sqrt(2)*lengthscales),2,-1)
    
    def parse_assign_param(self,
            pname, 
            param, 
            shape_param, 
            requires_grad_param,
            tfs_param,
            endsize_ops,
            constraints):
        return parse_assign_param(
            pname = pname,
            param = param, 
            shape_param = shape_param,
            requires_grad_param = requires_grad_param,
            tfs_param = tfs_param,
            endsize_ops = endsize_ops,
            constraints = constraints,
            torchify = self.torchify,
            npt = self.npt, 
            nptkwargs = self.nptkwargs)
    
class AbstractKernelScaleLengthscales(AbstractKernel):

    def __init__(self,
            d, 
            scale = 1.,
            lengthscales = 1.,
            shape_scale = [1],
            shape_lengthscales = None,
            tfs_scale = (tf_exp_eps_inv,tf_exp_eps),
            tfs_lengthscales = (tf_exp_eps_inv,tf_exp_eps),
            torchify = False,
            requires_grad_scale = True, 
            requires_grad_lengthscales = True, 
            device = "cpu",
            compile_call = False,
            comiple_call_kwargs = {},
            ):
        r"""
        Args:
            d (int): Dimension. 
            scale (Union[np.ndarray,torch.Tensor]): Scaling factor $S$.
            lengthscales (Union[np.ndarray,torch.Tensor]): Lengthscales $\boldsymbol{\gamma}$.
            shape_scale (list): Shape of `scale` when `np.isscalar(scale)`. 
            shape_lengthscales (list): Shape of `lengthscales` when `np.isscalar(lengthscales)`
            tfs_scale (Tuple[callable,callable]): The first argument transforms to the raw value to be optimized; the second applies the inverse transform.
            tfs_lengthscales (Tuple[callable,callable]): The first argument transforms to the raw value to be optimized; the second applies the inverse transform.
            torchify (bool): If `True`, use the `torch` backend. Set to `True` if computing gradients with respect to inputs and/or hyperparameters.
            requires_grad_scale (bool): If `True` and `torchify`, set `requires_grad=True` for `scale`.
            requires_grad_lengthscales (bool): If `True` and `torchify`, set `requires_grad=True` for `lengthscales`.
            device (torch.device): If `torchify`, put things onto this device.
            compile_call (bool): If `True`, `torch.compile` the `parsed___call__` method. 
            comiple_call_kwargs (dict): When `compile_call` is `True`, pass these keyword arguments to `torch.compile`.
        """
        super().__init__(d=d,torchify=torchify,device=device,compile_call=compile_call,comiple_call_kwargs=comiple_call_kwargs)
        self.raw_scale= self.parse_assign_param(
            pname = "scale",
            param = scale, 
            shape_param = shape_scale,
            requires_grad_param = requires_grad_scale,
            tfs_param = tfs_scale,
            endsize_ops = [1],
            constraints = ["POSITIVE"])
        self.tfs_scale = tfs_scale
        self.batch_param_names.append("scale")
        self.raw_lengthscales = self.parse_assign_param(
            pname = "lengthscales",
            param = lengthscales, 
            shape_param = [self.d] if shape_lengthscales is None else shape_lengthscales,
            requires_grad_param = requires_grad_lengthscales,
            tfs_param = tfs_lengthscales,
            endsize_ops = [1,self.d],
            constraints = ["POSITIVE"])
        self.tfs_lengthscales = tfs_lengthscales
        self.batch_param_names.append("lengthscales")

    @property
    def scale(self):
        return self.tfs_scale[1](self.raw_scale)
    
    @property
    def lengthscales(self):
        return self.tfs_lengthscales[1](self.raw_lengthscales)
        


from .abstract_stopping_criterion import AbstractStoppingCriterion
from .cub_mc_clt import CubMCCLT
from .cub_qmc_rep_student_t import CubQMCRepStudentT
from .cub_mc_g import CubMCG
from .cub_qmc_lattice_g import CubQMCLatticeG
from .cub_qmc_net_g import CubQMCNetG
from .cub_mlmc import CubMLMC
from .cub_mlqmc import CubMLQMC
from .cub_mlmc_cont import CubMLMCCont
from .cub_mlqmc_cont import CubMLQMCCont
from .cub_qmc_bayes_lattice_g import CubQMCBayesLatticeG
from .cub_qmc_bayes_net_g import CubQMCBayesNetG
from .cub_mc_clt_vec import CubMCCLTVec
try: 
    import torch
    import gpytorch
    from .pf_gp_ci import PFGPCI,PFSampleErrorDensityAR,SuggesterSimple
except ImportError:
    class PFGPCI(object):
        def __init__(self, *args, **kwargs):
            raise Exception("PFGPCI requires torch and gpytorch but no installations found")
    class PFSampleErrorDensityAR(object):
        def __init__(self, *args, **kwargs):
            raise Exception("PFSampleErrorDensityAR requires torch and gpytorch but no installations found")
    class SuggesterSimple(object):
        def __init__(self, *args, **kwargs):
            raise Exception("SuggesterSimple requires torch and gpytorch but no installations found")

StoppingCriterion = AbstractStoppingCriterion
_StoppingCriterion = AbstractStoppingCriterion
CubQMCRep = CubQMCRepStudentT
CubQMCCLT = CubQMCRepStudentT
CubQMCDigitalNetB2G = CubQMCNetG
CubQMCSobolG = CubQMCNetG
CubBayesLatticeG = CubQMCBayesLatticeG
CubBayesNetG = CubQMCBayesNetG
CubQMCCBayesDigitalNetB2G = CubQMCBayesNetG
CubBayesDigitalNetB2G = CubQMCBayesNetG
CubQMCBayesSobolG = CubQMCBayesNetG
CubBayesSobolG = CubQMCBayesNetG
CubQMCMLCont = CubMLQMCCont
CubQMCML = CubMLQMC
CubMCMLCont = CubMLMCCont
CubMCML = CubMLMC


import numpy as np
from .exceptions_warnings import ParameterError
from .torch_numpy_ops import get_npt

def k4sumterm(x, t, cutoff=1e-8):
    r""" 
    $$K_4(x) = \sum_{a=0}^{t-1} \frac{x_a}{2^{3a}}$$

    where $x_a$ is the bit at index $a$ in the binary expansion of $x$
    e.g. $x = 6$ with $t=3$ has $(x_0,x_1,x_2) = (1,1,0)$

    Examples:
        >>> t = 3
        >>> rng = np.random.Generator(np.random.SFC64(11))
        >>> x = rng.integers(low=0,high=2**t,size=(5,4))
        >>> with np.printoptions(precision=2):
        ...     k4sumterm(x,t)
        array([[ 1.11,  0.89,  1.11, -0.89],
               [ 1.11,  1.14,  1.11, -0.86],
               [-0.89,  0.86, -1.11,  0.89],
               [-1.11,  0.89, -0.89,  0.89],
               [-1.14, -0.89, -0.89, -0.86]])
        >>> import torch 
        >>> with torch._tensor_str.printoptions(precision=2):
        ...     k4sumterm(torch.from_numpy(x),t)
        tensor([[ 1.11,  0.89,  1.11, -0.89],
                [ 1.11,  1.14,  1.11, -0.86],
                [-0.89,  0.86, -1.11,  0.89],
                [-1.11,  0.89, -0.89,  0.89],
                [-1.14, -0.89, -0.89, -0.86]])
    
    Args:
        x (Union[np.ndarray torch.Tensor]): Integer arrays.
        t (int): Number of bits in each integer.
    
    Returns:
        y (Union[np.ndarray torch.Tensor]): The $K_4$ sum term.
    """
    total = 0.
    for a in range(0,t):
        factor = 1/float(2.**(3*a))
        if factor<cutoff: break
        total += (-1.)**((x>>(t-a-1))&1)*factor
    return total

WEIGHTEDWALSHFUNCSPOS = {
    2: lambda beta,xf,xb,t: -beta*xf + 5/2*(1-2**(-beta)),
    3: lambda beta,xf,xb,t: beta*xf**2 - 5*(1-2**(-beta))*xf + 43/18*(1-2**(-2*beta)),
    4: lambda beta,xf,xb,t: -2/3*beta*xf**3 + 5*(1-2**(-beta))*xf**2 - 43/9*(1-2**(-2*beta))*xf + 701/294*(1-2**(-3*beta)) + beta*(1/48*k4sumterm(xb,t)-1/42),
}

WEIGHTEDWALSHFUNCSZEROS = {
    2: 5/2,
    3: 43/18,
    4: 701/294,
}

def weighted_walsh_funcs(alpha, xb, t):
    r"""
    Weighted walsh functions 

    $$\sum_{k=0}^\infty \mathrm{wal}_k(x) 2^{-\mu_\alpha(k)}$$ 

    where $\mathrm{wal}_k$ is the $k^\text{th}$ Walsh function 
    and $\mu_\alpha$ is the Dick weight function which sums the first $\alpha$ largest indices of $1$ bits in the binary expansion of $k$ 
    e.g. $k=13=1101_2$ has 1-bit indexes $(4,3,1)$ so 
    
    $$\mu_1(k) = 4, \mu_2(k) = 4+3, \mu_3(k) = 4+3+1 = \mu_4(k) = \mu_5(k) = \dots$$

    Examples:
        >>> t = 3 
        >>> rng = np.random.Generator(np.random.SFC64(11))
        >>> xb = rng.integers(low=0,high=2**t,size=(2,3))
        >>> available_alpha = list(WEIGHTEDWALSHFUNCSPOS.keys())
        >>> available_alpha
        [2, 3, 4]
        >>> for alpha in available_alpha:
        ...     y = weighted_walsh_funcs(alpha,xb,t)
        ...     with np.printoptions(precision=2):
        ...         print("alpha = %d\n%s"%(alpha,y))
        alpha = 2
        [[1.81 1.38 1.81]
         [0.62 1.81 2.5 ]]
        alpha = 3
        [[1.85 1.43 1.85]
         [0.62 1.85 2.39]]
        alpha = 4
        [[1.85 1.43 1.85]
         [0.62 1.85 2.38]]
        >>> import torch
        >>> for alpha in available_alpha:
        ...     y = weighted_walsh_funcs(alpha,torch.from_numpy(xb),t)
        ...     with torch._tensor_str.printoptions(precision=2):
        ...         print("alpha = %d\n%s"%(alpha,y))
        alpha = 2
        tensor([[1.81, 1.38, 1.81],
                [0.62, 1.81, 2.50]])
        alpha = 3
        tensor([[1.85, 1.43, 1.85],
                [0.62, 1.85, 2.39]])
        alpha = 4
        tensor([[1.85, 1.43, 1.85],
                [0.62, 1.85, 2.38]])
    
    Args:
        alpha (int): Weighted walsh functions order.
        xb (Union[np.ndarray,torch.Tensor]): Jnteger points at which to evaluate the weighted Walsh function.
        t (int): Number of bits in each integer in xb.
    
    returns:
        y (Union[np.ndarray,torch.Tensor]): Weighted Walsh function values.
    
    **References:**
        
    1.  Dick, Josef.  
        "Walsh spaces containing smooth functions and quasi–Monte Carlo rules of arbitrary high order."  
        SIAM Journal on Numerical Analysis 46.3 (2008): 1519-1553.

    2.  Dick, Josef.  
        "The decay of the Walsh coefficients of smooth functions."  
        Bulletin of the Australian Mathematical Society 80.3 (2009): 430-453.    
    """
    assert isinstance(alpha,int)
    assert alpha in WEIGHTEDWALSHFUNCSPOS, "alpha = %d not in WEIGHTEDWALSHFUNCSPOS"%alpha
    assert alpha in WEIGHTEDWALSHFUNCSZEROS, "alpha = %d not in WEIGHTEDWALSHFUNCSZEROS"%alpha
    if isinstance(xb,np.ndarray):
        np_or_torch = np 
        y = np.ones(xb.shape) 
    else:
        import torch 
        np_or_torch = torch 
        y = torch.ones(xb.shape,device=xb.device)
    pidxs = xb>0
    y[~pidxs] = WEIGHTEDWALSHFUNCSZEROS[alpha]
    xfpidxs = (2**(-t))*xb[pidxs]
    betapidxs = -np_or_torch.floor(np_or_torch.log2(xfpidxs))
    y[pidxs] = WEIGHTEDWALSHFUNCSPOS[alpha](betapidxs,xfpidxs,xb[pidxs],t)
    return y

def to_bin(x, t):
    r"""
    Convert floating point representations of digital net samples in base $b=2$ to binary representations.

    Examples:
        >>> xf = np.random.Generator(np.random.PCG64(7)).uniform(low=0,high=1,size=(5))
        >>> xf 
        array([0.62509547, 0.8972138 , 0.77568569, 0.22520719, 0.30016628])
        >>> xb = to_bin(xf,2)
        >>> xb
        array([2, 3, 3, 0, 1], dtype=uint64)
        >>> to_bin(xb,2) 
        array([2, 3, 3, 0, 1], dtype=uint64)
        >>> import torch 
        >>> xftorch = torch.from_numpy(xf) 
        >>> xftorch
        tensor([0.6251, 0.8972, 0.7757, 0.2252, 0.3002], dtype=torch.float64)
        >>> xbtorch = to_bin(xftorch,2)
        >>> xbtorch
        tensor([2, 3, 3, 0, 1])
        >>> to_bin(xbtorch,2)
        tensor([2, 3, 3, 0, 1])

    
    Args:
        x (Union[np.ndarray,torch.Tensor]): floating point representation of samples. 
        t (int): number of bits in binary represtnations. Typically `dnb2.t` where `isinstance(dnb2,DigitalNetB2)`.
    
    Returns: 
        xb (Unioin[np.ndarray,torch.Tensor]): binary representation of samples with `dtype` either `np.uint64` or `torch.int64`. 
    """
    npt = get_npt(x)
    if npt==np:
        if npt.issubdtype(x.dtype,npt.floating):
            return npt.floor((x%1)*2.**t).astype(npt.uint64)
        elif npt.issubdtype(x.dtype,npt.integer):
            return x
        else:
            raise ParameterError("x.dtype must be float or int, got %s"%str(x.dtype))
    else: # npt==torch
        if npt.is_floating_point(x):
            return npt.floor((x%1)*2.**t).to(npt.int64)
        elif (not npt.is_floating_point(x)) and (not npt.is_complex(x)): # int type 
            return x 
        else :
            raise ParameterError("x.dtype must be float or int, got %s"%str(x.dtype))
    return xb

def to_float(x, t):
    r"""
    Convert binary representations of digital net samples in base $b=2$ to floating point representations.
    
    Examples:
        >>> xb = np.arange(8,dtype=np.uint64)
        >>> xb 
        array([0, 1, 2, 3, 4, 5, 6, 7], dtype=uint64)
        >>> to_float(xb,3)
        array([0.   , 0.125, 0.25 , 0.375, 0.5  , 0.625, 0.75 , 0.875])
        >>> xbtorch = bin_from_numpy_to_torch(xb)
        >>> xbtorch
        tensor([0, 1, 2, 3, 4, 5, 6, 7])
        >>> to_float(xbtorch,3)
        tensor([0.0000, 0.1250, 0.2500, 0.3750, 0.5000, 0.6250, 0.7500, 0.8750])
    
    Args:
        x (Union[np.ndarray,torch.Tensor]): binary representation of samples with `dtype` either `np.uint64` or `torch.int64`.
        t (int): number of bits in binary represtnations. Typically `dnb2.t` where `isinstance(dnb2,DigitalNetB2)`.
    
    Returns: 
        xf (Unioin[np.ndarray,torch.Tensor]): floating point representation of samples.  
    """
    npt = get_npt(x)
    if npt==np: # npt==torch
        if x.dtype==np.uint64: 
            return x.astype(np.float64)*2.**(-t)
        elif npt.is_floating_point(x):
            return x 
        else: 
            raise ParameterError("x.dtype must be np.uint64, got %s"%str(x.dtype))
    else:
        if x.dtype==npt.int64: 
            return x.to(npt.get_default_dtype())*2.**(-t) 
        elif npt.is_floating_point(x):
            return x 
        else:
            raise ParameterError("x.dtype must be torch.int64, got %s"%str(x.dtype))

def bin_from_numpy_to_torch(xb):
    r"""
    Convert `numpy.uint64` to `torch.int64`, useful for converting binary samples from `DigitalNetB2` to torch representations.
    
    Examples:
        >>> xb = np.arange(8,dtype=np.uint64)
        >>> xb 
        array([0, 1, 2, 3, 4, 5, 6, 7], dtype=uint64)
        >>> bin_from_numpy_to_torch(xb)
        tensor([0, 1, 2, 3, 4, 5, 6, 7])
    
    Args:
        xb (Union[np.ndarray]): binary representation of samples with `dtype=np.uint64`
    
    Returns: 
        xbtorch (Unioin[torch.Tensor]): binary representation of samples with `dtype=torch.int64`.  
    """
    assert xb.dtype==np.uint64
    assert xb.max()<=(2**63-1), "require all xb < 2^63"
    import torch
    return torch.from_numpy(xb.astype(np.int64))


import numpy as np
from .abstract_integrand import AbstractIntegrand
from ..discrete_distribution import DigitalNetB2
from ..true_measure import Uniform
from ..util import ParameterError

class Ishigami(AbstractIntegrand):
    r"""
    Ishigami function in $d=3$ dimensions from [1] and [https://www.sfu.ca/~ssurjano/ishigami.html](https://www.sfu.ca/~ssurjano/ishigami.html). 

    $$g(\boldsymbol{t}) = (1+bt_2^4)\sin(t_0)+a\sin^2(t_1), \qquad \boldsymbol{T} = (T_0,T_1,T_2) \sim \mathcal{U}(-\pi,\pi)^3.$$
    
    Examples:
        >>> integrand = Ishigami(DigitalNetB2(3,seed=7))
        >>> y = integrand(2**10)
        >>> y.shape 
        (1024,)
        >>> print("%.4f"%y.mean())
        3.5000
        >>> integrand.true_measure
        Uniform (AbstractTrueMeasure)
            lower_bound     -3.142
            upper_bound     3.142
        
        With independent replications

        >>> integrand = Ishigami(DigitalNetB2(3,seed=7,replications=2**4))
        >>> y = integrand(2**6)
        >>> y.shape
        (16, 64)
        >>> muhats = y.mean(-1) 
        >>> muhats.shape 
        (16,)
        >>> print("%.4f"%muhats.mean())
        3.4646
        
    **References:**

    1.  Ishigami, T., & Homma, T.  
        An importance quantification technique in uncertainty analysis for computer models.  
        In Uncertainty Modeling and Analysis, 1990.  
        Proceedings, First International Symposium on (pp. 398-403). IEEE.
    """
    def __init__(self,sampler, a=7, b=.1):
        r"""
        Args:
            sampler (Union[AbstractDiscreteDistribution,AbstractTrueMeasure]): Either  
                
                - a discrete distribution from which to transform samples, or
                - a true measure by which to compose a transform.
            a (float): First parameter $a$.
            b (float): Second parameter $b$.
        """
        self.sampler = sampler
        if self.sampler.d != 3:
            raise ParameterError("Ishigami integrand requires 3 dimensional sampler")
        self.a = a
        self.b = b
        self.true_measure = Uniform(self.sampler, lower_bound=-np.pi, upper_bound=np.pi)
        super(Ishigami,self).__init__(dimension_indv=(),dimension_comb=(),parallel=False)
    
    def g(self, t):
        y = (1+self.b*t[...,2]**4)*np.sin(t[...,0])+self.a*np.sin(t[...,1])**2
        return y

    def _spawn(self, level, sampler):
        return Ishigami(sampler=sampler,a=self.a,b=self.b)
    
    @staticmethod
    def _exact_sensitivity_indices(indices, a, b):
        a,b = np.atleast_1d(a),np.atleast_1d(b)
        assert a.shape==b.shape and a.ndim==1 and b.ndim==1
        mu = a/2
        m2 = 1/2+3/8*a**2+np.pi**4/5*b+np.pi**8/18*b**2
        tau_closed = {
            repr([True,False,False]): 1/50*(5+np.pi**4*b)**2,
            repr([False,True,False]): a**2/8,
            repr([False,False,True]): 0,
            repr([True,True,False]): a**2/8+1/50*(5+np.pi**4*b)**2,
            repr([True,False,True]): 1/90*(45+18*np.pi**4*b+5*np.pi**8*b**2),
            repr([False,True,True]): a**2/8}
        tau_total = {
            repr([False,True,True]): a**2/8+8*np.pi**8/225*b**2,
            repr([True,False,True]): 1/90*(45+18*np.pi**4*b+5*np.pi**8*b**2),
            repr([True,True,False]): 1/2+a**2/8+np.pi**4/5*b+np.pi**8/18*b**2,
            repr([False,False,True]): 8*np.pi**8/225*b**2,
            repr([False,True,False]): a**2/8,
            repr([True,False,False]): 1/90*(45+18*np.pi**4*b+5*np.pi**8*b**2)}
        solution = np.zeros((2,len(indices),len(a)),dtype=float)
        for j,idx in enumerate(indices):
            solution[0,j] = tau_closed[repr(idx.tolist())]/(m2-mu**2)
            solution[1,j] = tau_total[repr(idx.tolist())]/(m2-mu**2)
        return solution.squeeze()
        
    @staticmethod
    def _exact_fu_functions(x,indices,a,b):
        x = np.atleast_2d(x)
        n = len(x)
        a,b = np.atleast_1d(a),np.atleast_1d(b)
        assert x.ndim==2 and x.shape==(n,3) and a.shape==(1,) and b.shape==(1,)
        x0,x1,x2 = x[:,0],x[:,1],x[:,2]
        fus = {
            repr([]): a/2,
            repr([0]): -1/5*(5+np.pi**4*b)*np.sin(2*np.pi*x0),
            repr([1]): -1/2*a*np.cos(4*np.pi*x1),
            repr([2]): 0,
            repr([0,1]): 0,
            repr([0,2]): -4/5*b*np.pi**4*(1+10*(-1+x2)*x2*(1+2*(-1+x2)*x2))*np.sin(2*np.pi*x0),
            repr([1,2]): -1/2*a*np.cos(4*np.pi*x1)+1/5*(5+b*np.pi**4)*np.sin(2*np.pi*x2),
            repr([0,1,2]): 1/2*a*np.cos(4*np.pi*x1)-1/5*(5+b*np.pi**4)*np.sin(2*np.pi*x1)}
        solution = np.zeros((n,len(indices)),dtype=float)
        for j,idx in enumerate(indices): solution[:,j] = fus[repr(idx)]
        return solution.squeeze()



from ..util import MethodImplementationError, _univ_repr, DimensionError, ParameterError
from ..discrete_distribution.abstract_discrete_distribution import AbstractDiscreteDistribution
import numpy as np
from typing import Union


class AbstractTrueMeasure(object):

    def __init__(self):
        prefix = 'A concrete implementation of TrueMeasure must have '
        if not hasattr(self,'domain'):
            raise ParameterError(prefix + 'self.domain, 2xd ndarray of domain lower bounds (first col) and upper bounds (second col)')
        if not hasattr(self,'range'):
            raise ParameterError(prefix + 'self.range, 2xd ndarray of range lower bounds (first col) and upper bounds (second col)')
        if not hasattr(self,'parameters'):
            self.parameters = []
    
    def _parse_sampler(self, sampler):
        self.sub_compatibility_error = False
        if isinstance(sampler,AbstractDiscreteDistribution):
            self.transform = self # this is the initial transformation, \Psi_0
            self.d = sampler.d # take the dimension from the discrete distribution
            self.discrete_distrib = sampler
            if sampler.mimics == 'StdUniform':
                if not (self.domain==np.tile([0,1],(self.d,1))).all():
                    raise ParameterError("The True measure's transform should have unit-cube domain.")
            else:
                raise ParameterError('The %s true measure does not support discrete distributions that mimic a %s.'%(type(self).__name__,sampler.mimics))
        elif isinstance(sampler,AbstractTrueMeasure):
            self.transform = sampler # this is a composed transform, \Psi_j for j>0
            self.parameters += ['transform']
            self.d = sampler.d # take the dimension from the sub-sampler (composed transform)
            self.discrete_distrib = self.transform.discrete_distrib
            if (self.domain!=self.transform.range).any():
                self.sub_compatibility_error = True
            if self.transform.sub_compatibility_error:
                raise ParameterError("The sub-transform domain must match the sub-sub-transform range.")
        else:
            raise ParameterError("sampler input should either be a AbstractDiscreteDistribution or AbstractTrueMeasure")
    
    def __call__(self, n=None, n_min=None, n_max=None, return_weights=False, warn=True):
        r"""
        - If just `n` is supplied, generate samples from the sequence at indices 0,...,`n`-1.
        - If `n_min` and `n_max` are supplied, generate samples from the sequence at indices `n_min`,...,`n_max`-1.
        - If `n` and `n_min` are supplied, then generate samples from the sequence at indices `n`,...,`n_min`-1.

        Args:
            n (Union[None,int]): Number of points to generate.
            n_min (Union[None,int]): Starting index of sequence.
            n_max (Union[None,int]): Final index of sequence.
            return_weights (bool): If `True`, return `weights` as well
            warn (bool): If `False`, disable warnings when generating samples.

        Returns:
            t (np.ndarray): Samples from the sequence. 
                
                - If `replications` is `None` then this will be of size (`n_max`-`n_min`) $\times$ `dimension` 
                - If `replications` is a positive int, then `t` will be of size `replications` $\times$ (`n_max`-`n_min`) $\times$ `dimension` 
            weights (np.ndarray): Only returned when `return_weights=True`. The Jacobian weights for the transformation
        """
        return self.gen_samples(n=n,n_min=n_min,n_max=n_max,return_weights=return_weights,warn=warn)
    
    def gen_samples(self, n=None, n_min=None, n_max=None, return_weights=False, warn=True):
        x = self.discrete_distrib(n=n,n_min=n_min,n_max=n_max,warn=warn)
        assert isinstance(return_weights,bool)
        return self._jacobian_transform_r(x=x,return_weights=return_weights)
        
    def _jacobian_transform_r(self, x, return_weights):
        r""" Recursive Jacobian transform. """
        if self.sub_compatibility_error:
            raise ParameterError("The transform domain must match the sub-transform range.")
        if self.transform == self: # is \Psi_0
            if return_weights:
                t = self._transform(x)
                jac = 1/self._weight(t)
            else:
                t = self._transform(x)
        else: # is transform \Psi_j for j>0
            if return_weights:
                t_sub,jac_sub = self.transform._jacobian_transform_r(x=x,return_weights=return_weights)
                t = self._transform(t_sub)
                jac = jac_sub/self._weight(t) # |\Psi1/\lambda(\psi(x))
            else:
                t_sub = self.transform._jacobian_transform_r(x=x,return_weights=return_weights)
                t = self._transform(t_sub)
        if return_weights:
            return t,jac 
        else:
            return t
        
    def _transform(self, x): 
        r""" Transformation from the standard uniform to the true measure distribution. """
        raise MethodImplementationError(self,'_transform. Try setting sampler to be in a PDF AbstractTrueMeasure to importance sample by.')
   
    def _weight(self, x):
        r"""
        Non-negative weight function.  
        This is often a PDF, but is not required to be  
        e.g., Lebesgue weight is always 1, but is not a PDF.

        Args:
            x (np.ndarray): n x d  matrix of samples
        
        Returns:
            np.ndarray: length n vector of weights at locations of x
        """ 
        raise MethodImplementationError(self,'weight. Try a different true measure with a _weight method.') 
     
    def spawn(self, s=1, dimensions=None):
        r"""
        Spawn new instances of the current true measure but with new seeds and dimensions.
        Used by multi-level QMC algorithms which require different seeds and dimensions on each level.

        Note:
            Use `replications` instead of using `spawn` when possible, e.g., when spawning copies which all have the same dimension.

        Args:
            s (int): Number of copies to spawn
            dimensions (np.ndarray): Length `s` array of dimensions for each copy. Defaults to the current dimension. 

        Returns:
            spawned_true_measures (list): True measure with new seeds and dimensions.
        """
        sampler = self.discrete_distrib if self.transform==self else self.transform
        sampler_spawns = sampler.spawn(s=s,dimensions=dimensions)
        spawned_true_measures = [None]*len(sampler_spawns)
        for i in range(s):
            spawned_true_measures[i] = self._spawn(sampler_spawns[i],sampler_spawns[i].d)
        return spawned_true_measures
    
    def _spawn(self, sampler, dimension):
        raise MethodImplementationError(self, '_spawn')  
        
    def __repr__(self):
        return _univ_repr(self, "AbstractTrueMeasure", self.parameters)
    




from .abstract_discrete_distribution import AbstractIIDDiscreteDistribution
from ..util import ParameterError, ParameterWarning
import numpy as np
from typing import Union
import warnings

class IIDStdUniform(AbstractIIDDiscreteDistribution):
    r"""
    IID standard uniform points, a wrapper around [`numpy.random.rand`](https://numpy.org/doc/stable/reference/random/generated/numpy.random.rand.html). 

    Note:
        - Unlike low discrepancy sequence, calling an `IIDStdUniform` instance gives new samples every time,  
            e.g., running the first doctest below with `dd = Lattice(dimension=2)` would give the same 4 points in both calls,  
            but since we are using an `IIDStdUniform` instance it gives different points every call. 
    
    Examples:
        >>> discrete_distrib = IIDStdUniform(dimension=2,seed=7)
        >>> discrete_distrib(4)
        array([[0.04386058, 0.58727432],
               [0.3691824 , 0.65212985],
               [0.69669968, 0.10605352],
               [0.63025643, 0.13630282]])
        >>> discrete_distrib(4) # gives new samples every time
        array([[0.5968363 , 0.0576251 ],
               [0.2028797 , 0.22909681],
               [0.1366783 , 0.75220658],
               [0.84501765, 0.56269008]])
        >>> discrete_distrib
        IIDStdUniform (AbstractIIDDiscreteDistribution)
            d               2^(1)
            replications    1
            entropy         7

        Replications (implemented for API consistency) 


        >>> x = IIDStdUniform(dimension=3,replications=2,seed=7)(4)
        >>> x.shape
        (2, 4, 3)
        >>> x
        array([[[0.04386058, 0.58727432, 0.3691824 ],
                [0.65212985, 0.69669968, 0.10605352],
                [0.63025643, 0.13630282, 0.5968363 ],
                [0.0576251 , 0.2028797 , 0.22909681]],
        <BLANKLINE>
               [[0.1366783 , 0.75220658, 0.84501765],
                [0.56269008, 0.04826852, 0.71308655],
                [0.80983568, 0.85383675, 0.80475135],
                [0.6171181 , 0.1239209 , 0.16809479]]])
    """

    def __init__(self,
                 dimension = 1, 
                 replications = None, 
                 seed = None):
        r"""
        Args:
            dimension (int): Dimension of the samples.
            replications (Union[None,int]): Number of randomizations. This is implemented only for API consistency. Equivalent to reshaping samples. 
            seed (Union[None,int,np.random.SeedSeq): Seed the random number generator for reproducibility.
        """
        super(IIDStdUniform,self).__init__(int(dimension),replications,seed,d_limit=np.inf,n_limit=np.inf)
        if not (self.dvec==np.arange(self.d)).all():
            warnings.warn("IIDStdUniform does not accomodate dvec",ParameterWarning)

    def _gen_samples(self, n_min, n_max, return_binary, warn):
        if n_min>0 and warn:
            raise warnings.warn("For IIDStdUniform setting n_min>0 makes no difference as new samples are generated every call regardless, e.g., calling gen_samples(n_min=0,n_max=4) twice will give different samples.",ParameterWarning)
        if return_binary:
            raise ParameterError("IIDStdUniform does not support return_binary=True")
        x = self.rng.uniform(size=(self.replications,int(n_max-n_min),self.d))
        return x
        
    def _spawn(self, child_seed, dimension):
        return IIDStdUniform(
            dimension = dimension,
            replications = None if self.no_replications else self.replications,
            seed = child_seed,
        )



from ..integrand.abstract_integrand import AbstractIntegrand
from ..util import DistributionCompatibilityError, ParameterError, MethodImplementationError, _univ_repr
import numpy as np


class AbstractStoppingCriterion(object):
    
    def __init__(self, allowed_distribs, allow_vectorized_integrals):
        """
        Args:
            allowed_distribs (list): Allowed discrete distribution classes.
            allow_vectorized_integrals (bool): Whether or not to allow integrands with vectorized outputs, 
                i.e., those with `integrand.d_indv!=()`. 
        """
        sname = type(self).__name__
        prefix = 'A concrete implementation of StoppingCriterion must have '
        # integrand check
        if (not hasattr(self, 'integrand')) or (not isinstance(self.integrand,AbstractIntegrand)):
            raise ParameterError(prefix + 'self.integrand, an Integrand instance')
        # true measure check
        if (not hasattr(self, 'true_measure')) or (self.true_measure!=self.integrand.true_measure):
            raise ParameterError(prefix + 'self.true_measure=self.integrand.true_measure')
        # discrete distribution check
        if (not hasattr(self, 'discrete_distrib')) or (self.discrete_distrib!=self.integrand.discrete_distrib):
            raise ParameterError(prefix + 'self.discrete_distrib=self.integrand.discrete_distrib')
        if not isinstance(self.discrete_distrib,tuple(allowed_distribs)):
            raise DistributionCompatibilityError('%s must have a DiscreteDistribution in %s'%(sname,str(allowed_distribs)))
        if (not allow_vectorized_integrals) and len(self.integrand.d_indv)>0:
            raise ParameterError('Vectorized integrals (with d_indv!=() outputs per sample) are not supported by this stopping criterion')
        # parameter checks
        if not hasattr(self,'parameters'):
            self.parameters = []
            
    def integrate(self):
        """
        *Abstract method* to determine the number of samples needed to satisfy the tolerance(s).

        Returns:
            solution (Union[float,np.ndarray]): Approximation to the integral with shape `integrand.d_comb`.
            data (Data): A data object.
        """
        raise MethodImplementationError(self, 'integrate')
    
    def _compute_indv_alphas(self, alphas_comb):
        alphas_indv = np.tile(1,self.integrand.d_indv)
        identity_dependency = True
        for k in np.ndindex(self.integrand.d_comb):
            comb_flags = np.tile(True,self.integrand.d_comb)
            comb_flags[k] = False
            flags_indv = self.integrand.dependency(comb_flags)
            if self.integrand.d_indv!=self.integrand.d_comb or (flags_indv!=comb_flags).any(): identity_dependency=False
            dependents_k = ~flags_indv
            n_dep_k = dependents_k.sum()
            alpha_k = alphas_comb[k]/n_dep_k
            alpha_k_mat = alpha_k*dependents_k
            alphas_indv = np.where(alpha_k_mat==0,alphas_indv,np.minimum(alpha_k_mat,alphas_indv))
        return alphas_indv,identity_dependency
    
    def set_tolerance(self, abs_tol=None, rel_tol=None, rmse_tol=None):
        """
        Reset the tolerances.
        
        Args:
            abs_tol (float): Absolute tolerance (if applicable). Reset if supplied, ignored otherwise. 
            rel_tol (float): Relative tolerance (if applicable). Reset if supplied, ignored otherwise. 
            rmse_tol (float): RMSE tolerance (if applicable). Reset if supplied, ignored if not. 
                If `rmse_tol` is not supplied but `abs_tol` is, then `rmse_tol = abs_tol / norm.ppf(1-alpha/2)`. 
        """
        raise MethodImplementationError(self, 'integrate')

    def __repr__(self):
        return _univ_repr(self, "AbstractStoppingCriterion", self.parameters)



from .abstract_true_measure import AbstractTrueMeasure
from ..util import DimensionError, ParameterError
from ..discrete_distribution import DigitalNetB2
import numpy as np
from typing import Union

class BernoulliCont(AbstractTrueMeasure):
    r"""
    Continuous Bernoulli distribution with independent marginals as described in [https://en.wikipedia.org/wiki/Continuous_Bernoulli_distribution](https://en.wikipedia.org/wiki/Continuous_Bernoulli_distribution).

    Examples: 
        >>> true_measure = BernoulliCont(DigitalNetB2(2,seed=7),lam=.2)
        >>> true_measure(4)
        array([[0.56205914, 0.83607872],
               [0.09433983, 0.28057299],
               [0.97190779, 0.01883497],
               [0.28050753, 0.39178506]])
        >>> true_measure
        BernoulliCont (AbstractTrueMeasure)
            lam             0.200
        
        With independent replications 

        >>> x = BernoulliCont(DigitalNetB2(3,seed=7,replications=2),lam=[.25,.5,.75])(4)
        >>> x.shape 
        (2, 4, 3)
        >>> x
        array([[[0.16343492, 0.1821862 , 0.83209443],
                [0.55140696, 0.66169442, 0.56381501],
                [0.35229402, 0.79818233, 0.13825119],
                [0.85773226, 0.29520621, 0.85203898]],
        <BLANKLINE>
               [[0.32359293, 0.85899604, 0.63591945],
                [0.40278251, 0.04353443, 0.46749989],
                [0.1530438 , 0.29281506, 0.116725  ],
                [0.6345258 , 0.60241448, 0.84822692]]])
    """

    def __init__(self, sampler, lam=1/2):
        r"""
        Args:
            sampler (Union[AbstractDiscreteDistribution,AbstractTrueMeasure]): Either  
                
                - a discrete distribution from which to transform samples, or
                - a true measure by which to compose a transform.
            lam (Union[float,np.ndarray]): Vector of shape parameters, each in $(0,1)$.
        """
        self.parameters = ['lam']
        self.domain = np.array([[0,1]])
        self.range = np.array([[0,1]])
        self._parse_sampler(sampler)
        self.lam = lam
        self.l = np.array(lam)
        if self.l.size==1:
            self.l = self.l.item()*np.ones(self.d)
        if not (self.l.shape==(self.d,) and (0<=self.l).all() and (self.l<=1).all()):
            raise DimensionError('lam must be scalar or have length equal to dimension and must be in (0,1).')
        super(BernoulliCont,self).__init__()

    def _transform(self, x):
        tf = np.zeros(x.shape,dtype=float)
        for j in range(self.d):
            tf[...,j] = x[...,j] if self.l[j]==1/2 else np.log(((2*self.l[j]-1)*x[...,j]-self.l[j]+1)/(1-self.l[j])) / np.log(self.l[j]/(1-self.l[j]))
        return tf
    
    def _weight(self, x):
        w = np.zeros(x.shape,dtype=float)
        for j in range(self.d):
            C = 2 if self.l[j]==1/2 else 2*np.arctanh(1-2*self.l[j])/(1-2*self.l[j])
            w[...,j] = C*self.l[j]**x[...,j]*(1-self.l[j])**(1-x[...,j])
        return np.prod(w,-1)
    
    def _spawn(self, sampler, dimension):
        if dimension==self.d: # don't do anything if the dimension doesn't change
            spawn = BernoulliCont(sampler,lam=self.lam)
        else:
            l0 = self.l[0]
            if (self.l!=l0).any():
                raise DimensionError('''
                        In order to spawn a BernoulliCont measure
                        lam must all be the same.''')
            spawn = BernoulliCont(sampler,lam=l0)
        return spawn



from .brownian_motion import BrownianMotion
from ..discrete_distribution import DigitalNetB2
from ..util import ParameterError
from numpy import exp, zeros, minimum, array, sqrt, log, pi, linalg, eye
from scipy.stats import multivariate_normal


class GeometricBrownianMotion(BrownianMotion):
    r"""
    A Geometric Brownian Motion (GBM) with initial value $S_0$, drift $\gamma$, and diffusion $\sigma^2$ is 

    $$\mathrm{GBM}(t) = S_0 \exp[(\gamma - \sigma^2/2) t + \sigma \mathrm{BM}(t)]$$

    where BM is a Brownian Motion drift $\gamma$ and diffusion $\sigma^2$. 
    
    Examples:
        >>> gbm = GeometricBrownianMotion(DigitalNetB2(4,seed=7), t_final=2, drift=0.1, diffusion=0.2)
        >>> gbm.gen_samples(2)
        array([[0.92343761, 1.42069027, 1.30851806, 0.99133819],
               [0.7185916 , 0.42028013, 0.42080335, 0.4696196 ]])
        >>> gbm
        GeometricBrownianMotion (AbstractTrueMeasure)
            time_vec        [0.5 1.  1.5 2. ]
            drift           0.100
            diffusion       0.200
            mean_gbm        [1.051 1.105 1.162 1.221]
            covariance_gbm  [[0.116 0.122 0.128 0.135]
                             [0.122 0.27  0.284 0.299]
                             [0.128 0.284 0.472 0.496]
                             [0.135 0.299 0.496 0.734]]
            decomp_type     PCA
    """

    def __init__(self, sampler, t_final=1, initial_value=1, drift=0, diffusion=1, decomp_type='PCA'):
        r"""
        Args:
            sampler (DiscreteDistribution/TrueMeasure): A discrete distribution or true measure.
            t_final (float): End time for the geometric Brownian motion, non-negative.
            initial_value (float): Positive initial value of the process, $S_0$.
            drift (float): Drift coefficient $\gamma$.
            diffusion (float): Positive diffusion coefficient $\sigma^2$.
            decomp_type (str): Method of decomposition, either "PCA" or "Cholesky".
        """
        super().__init__(sampler, t_final=t_final, drift=0, diffusion=diffusion, decomp_type=decomp_type)
        self.parameters = ['time_vec', 'drift', 'diffusion', 'mean_gbm', 'covariance_gbm', 'decomp_type']
        self.initial_value = initial_value
        self.drift = drift
        self.diffusion = diffusion
        self.mean_gbm = self._compute_gbm_mean()
        self.covariance_gbm = self._compute_gbm_covariance()
        self._setup_lognormal_distribution()
        self._validate_input()

    def _compute_gbm_mean(self):
        return self.initial_value * exp(self.drift * self.time_vec)

    def _compute_gbm_covariance(self):
        S0 = self.initial_value
        mu = self.drift
        t = array(self.time_vec) 
        # Vectorization using broadcasting
        t_sum = t[:, None] + t[None, :]  # t[i] + t[j] for all i,j
        t_min = minimum.outer(t, t)      # min(t[i], t[j]) for all i,j
        cov_matrix = (S0 ** 2) * exp(mu * t_sum) * (exp(self.diffusion * t_min) - 1)
        
        return cov_matrix
        

    def _transform(self, x):
        bm_samples = super()._transform(x)
        # Note: bm_samples already includes the diffusion scaling from the parent BrownianMotion class
        # So we don't multiply by self.diffusion again here
        exponent = (self.drift - 0.5 * self.diffusion) * self.time_vec + bm_samples
        samples = self.initial_value * exp(exponent)
        return samples
  

    def _spawn(self, sampler, dimension):
        return GeometricBrownianMotion(
            sampler, t_final=self.t, initial_value=self.initial_value,
            drift=self.drift, diffusion=self.diffusion, decomp_type=self.decomp_type
        )

    def _validate_input(self):
        """
        Validates the input parameters of the GeometricBrownianMotion class.

        Raises:
            ValueError: If the end time `t_final' is negative.
            ValueError: If the diffusion coefficient is less than or equal to zero.
            ValueError: If the initial value is less than or equal to zero.
            ParameterError: If the decomposition type is not 'PCA' or 'Cholesky'.
        """
        if self.t < 0:
            raise ValueError(f"End time 't_final' must be non-negative. It should not be {self.t}.")
        if self.diffusion <= 0:
            raise ValueError(f"Diffusion coefficient must be positive. It should not be {self.diffusion}.")
        if self.initial_value <= 0:
            raise ValueError(f"Initial value must be positive. It should not be {self.initial_value}.")
        if self.decomp_type.upper() not in ['PCA', 'CHOLESKY']:
            raise ParameterError(f"Decomposition type must be 'PCA' or 'Cholesky'. It should not be {self.decomp_type}.")

    def _validate_samples(self, samples, strict=False):
        """
        Validate that generated GBM samples meet mathematical requirements.
        """
        min_val = samples.min()
        max_val = samples.max()
        num_negative = (samples <= 0).sum()
        total_samples = samples.size
        
        validation_results = {
            'is_valid': num_negative == 0,
            'min_value': min_val,
            'max_value': max_val,
            'num_negative': num_negative,
            'total_samples': total_samples,
            'fraction_negative': num_negative / total_samples if total_samples > 0 else 0
        }
        
        if num_negative > 0:
            error_msg = (f"GeometricBrownianMotion validation FAILED: "
                        f"{num_negative}/{total_samples} ({100*validation_results['fraction_negative']:.2f}%) "
                        f"samples are non-positive (min: {min_val:.6f}). "
                        f"This violates the mathematical definition of GBM where S(t) = S0 * exp(...) > 0.")
            
            if strict:
                raise ValueError(error_msg)
            else:
                import warnings
                warnings.warn(error_msg)
        
        return validation_results

    def _setup_lognormal_distribution(self):
        """Setup scipy multivariate normal for the log-transformed variables."""
        # Mean of log(S(t)/S0): (drift - 0.5*diffusion) * t
        log_mean = (self.drift - 0.5 * self.diffusion) * self.time_vec
        
        # Covariance of log returns: diffusion * min(t_i, t_j)
        time_matrix = minimum.outer(self.time_vec, self.time_vec)
        log_cov = self.diffusion * time_matrix
        
        self.log_mvn_scipy = multivariate_normal(mean=log_mean, cov=log_cov, allow_singular=True)

    def _weight(self, x):
        """
        Compute PDF of multivariate log-normal distribution.
        For log-normal: f(x) = (1/∏x_i) * φ(log(x/S0)) where φ is multivariate normal PDF.
        
        Args:
            x (ndarray): GBM sample paths of shape (n_samples, n_timepoints)
            
        Returns:
            ndarray: PDF values for each sample path
        """
        if x.ndim == 1:
            x = x.reshape(1, -1)
        
        # Transform to log space: log(x/S0)
        log_transformed = log(x / self.initial_value)
        
        # Get PDF of underlying multivariate normal
        normal_pdf = self.log_mvn_scipy.pdf(log_transformed)
        
        # Apply Jacobian transformation: divide by product of x values
        jacobian = 1.0 / x.prod(axis=1)
        
        return normal_pdf * jacobian



from .gaussian import Gaussian
from ..discrete_distribution.abstract_discrete_distribution import AbstractDiscreteDistribution
from ..discrete_distribution import DigitalNetB2
from .abstract_true_measure import AbstractTrueMeasure
from ..util import ParameterError, _univ_repr
import numpy as np
from typing import Union


class BrownianMotion(Gaussian):
    r"""
    Brownian Motion as described in [https://en.wikipedia.org/wiki/Brownian_motion](https://en.wikipedia.org/wiki/Brownian_motion).  
    For a standard Brownian Motion $W$ we define the Brownian Motion $B$ with initial value $B_0$, drift $\gamma$, and diffusion $\sigma^2$ to be

    $$B(t) = B_0 + \gamma t + \sigma W(t).$$
    
    Examples:
        >>> true_measure = BrownianMotion(DigitalNetB2(4,seed=7),t_final=2,drift=2)
        >>> true_measure(2)
        array([[0.82189263, 2.7851793 , 3.60126805, 3.98054724],
               [0.2610643 , 0.06170064, 1.06448269, 2.30990767]])
        >>> true_measure
        BrownianMotion (AbstractTrueMeasure)
            time_vec        [0.5 1.  1.5 2. ]
            drift           2^(1)
            mean            [1. 2. 3. 4.]
            covariance      [[0.5 0.5 0.5 0.5]
                             [0.5 1.  1.  1. ]
                             [0.5 1.  1.5 1.5]
                             [0.5 1.  1.5 2. ]]
            decomp_type     PCA
    
        With independent replications 

        >>> x = BrownianMotion(DigitalNetB2(3,seed=7,replications=2),t_final=2,drift=2)(4)
        >>> x.shape 
        (2, 4, 3)
        >>> x
        array([[[0.66154685, 1.50620966, 3.52322901],
                [1.77064217, 3.32782204, 4.45013223],
                [1.33558688, 3.26017547, 3.40692337],
                [2.10317345, 3.78961839, 6.17948096]],
        <BLANKLINE>
               [[1.77868019, 2.75347902, 3.41161419],
                [0.44891984, 2.53987304, 4.7224811 ],
                [0.23147948, 2.25289769, 3.00039101],
                [2.06762574, 3.21756319, 4.93375923]]])
    """

    def __init__(self, sampler, t_final=1, initial_value=0, drift=0, diffusion=1, decomp_type='PCA'):
        r"""
        Args:
            sampler (Union[AbstractDiscreteDistribution,AbstractTrueMeasure]): Either  
                
                - a discrete distribution from which to transform samples, or
                - a true measure by which to compose a transform.
            t_final (float): End time. 
            initial_value (float): Initial value $B_0$. 
            drift (int): Drift $\gamma$. 
            diffusion (int): Diffusion $\sigma^2$. 
            decomp_type (str): Method for decomposition for covariance matrix. Options include
             
                - `'PCA'` for principal component analysis, or 
                - `'Cholesky'` for cholesky decomposition.
        """
        self.parameters = ['time_vec', 'drift', 'mean', 'covariance', 'decomp_type']
        # default to transform from standard uniform
        self.domain = np.array([[0,1]])
        self._parse_sampler(sampler)
        self.t = t_final # exercise time
        self.initial_value = initial_value
        self.drift = drift
        self.diffusion = diffusion
        self.time_vec = np.linspace(self.t/self.d,self.t,self.d) # evenly spaced
        self.diffused_sigma_bm = self.diffusion*np.array([[min(self.time_vec[i],self.time_vec[j]) for i in range(self.d)] for j in range(self.d)])
        self.drift_time_vec_plus_init = self.drift*self.time_vec+self.initial_value # mean
        self._parse_gaussian_params(self.drift_time_vec_plus_init,self.diffused_sigma_bm,decomp_type)
        self.range = np.array([[-np.inf,np.inf]])
        super(Gaussian,self).__init__()
    
    def _spawn(self, sampler, dimension):
        return BrownianMotion(sampler,t_final=self.t,drift=self.drift,decomp_type=self.decomp_type)


from .abstract_integrand import AbstractIntegrand
from ..discrete_distribution import DigitalNetB2
from ..true_measure import Uniform
from ..util import ParameterError
import numpy as np

class Genz(AbstractIntegrand):
    r"""
    Genz function following the [`DAKOTA` implementation](https://snl-dakota.github.io/docs/6.17.0/users/usingdakota/examples/additionalexamples.html?highlight=genz#genz-functions). 

    $$g_\mathrm{oscillatory}(\boldsymbol{t}) = \cos\left(-\sum_{j=1}^d c_j t_j\right)$$ 

    or 

    $$g_\mathrm{corner-peak}(\boldsymbol{t}) = \left(1+\sum_{j=1}^d c_j t_j\right)^{-(d+1)}$$ 

    where 
    
    $$\boldsymbol{T} \sim \mathcal{U}[0,1]^d$$

    and the coefficients $\boldsymbol{c}$ are have three kinds 

    $$c_k^{(1)} = \frac{k-1/2}{d}, \qquad c_k^{(2)} = \frac{1}{k^2}, \qquad c_k^{(3)} = \exp\left(\frac{k \log(10^{-8})}{d}\right), \qquad k=1,\dots,d.$$

    Examples:
        >>> for kind_func in ['OSCILLATORY','CORNER PEAK']:
        ...     for kind_coeff in [1,2,3]:
        ...         integrand = Genz(DigitalNetB2(2,seed=7),kind_func=kind_func,kind_coeff=kind_coeff)
        ...         y = integrand(2**14)
        ...         mu_hat = y.mean()
        ...         print('%-15s %-3d %.3f'%(kind_func,kind_coeff,mu_hat))
        OSCILLATORY     1   -0.351
        OSCILLATORY     2   -0.329
        OSCILLATORY     3   -0.217
        CORNER PEAK     1   0.713
        CORNER PEAK     2   0.714
        CORNER PEAK     3   0.720

        With independent replications

        >>> integrand = Genz(DigitalNetB2(2,seed=7,replications=2**4),kind_func="CORNER PEAK",kind_coeff=3)
        >>> y = integrand(2**6)
        >>> y.shape
        (16, 64)
        >>> muhats = y.mean(-1) 
        >>> muhats.shape 
        (16,)
        >>> print("%.4f"%muhats.mean())
        0.7200
    """

    def __init__(self, sampler, kind_func='OSCILLATORY', kind_coeff=1):
        """
        Args:
            sampler (Union[AbstractDiscreteDistribution,AbstractTrueMeasure]): Either  
                
                - a discrete distribution from which to transform samples, or
                - a true measure by which to compose a transform.
            kind_func (str): Either `'OSCILLATORY'` or `'CORNER PEAK'`
            kind_coeff (int): 1, 2, or 3 for choice of coefficients 
        """
        self.kind_func = str(kind_func).upper().strip().replace("_"," ").replace("-"," ")
        self.kind_coeff = kind_coeff
        if (self.kind_func not in ['OSCILLATORY','CORNER PEAK']) or (self.kind_coeff not in [1,2,3]):
            raise ParameterError('''
                Genz expects 
                    kind_func in ['OSCILLATORY','CORNER PEAK'] and 
                    kind_coeffs in [1,2,3]''')
        self.sampler = sampler
        self.true_measure = Uniform(self.sampler)
        self.d = self.true_measure.d
        if self.kind_coeff==1: self.c = (np.arange(1,self.d+1)-.5)/self.d
        elif self.kind_coeff==2: self.c = 1/np.arange(1,self.d+1)**2
        elif self.kind_coeff==3: self.c = np.exp(np.arange(1,self.d+1)*np.log(10**(-8))/self.d)
        if self.kind_func=='OSCILLATORY':
            self.g = self.g_oscillatory
            self.c = 4.5*self.c/self.c.sum()
        elif self.kind_func=='CORNER PEAK':
            self.g = self.g_corner_peak
            self.c = 0.25*self.c/self.c.sum()
        self.c = self.c[None,:]
        self.parameters = ['kind_func','kind_coeff']
        super(Genz,self).__init__(dimension_indv=(),dimension_comb=(),parallel=False)
    
    def g_oscillatory(self, t):
        return np.cos(-(self.c*t).sum(-1))

    def g_corner_peak(self, t):
        return (1+(self.c*t).sum(-1))**(-(self.d+1))
    
    def _spawn(self, level, sampler):
        return Genz(sampler=sampler,kind_func=self.kinda_func,kind_coeff=self.kind_coeff)



from .abstract_integrand import AbstractIntegrand
from ..discrete_distribution import DigitalNetB2
from ..true_measure import Uniform
from ..util import ParameterError
import numpy as np
import os

class UMBridgeWrapper(AbstractIntegrand):
    """
    Wrapper around a [`UM-Bridge`](https://um-bridge-benchmarks.readthedocs.io/en/docs/index.html) model. See also the [`UM-Bridge` documentation for the QMCPy client](https://um-bridge-benchmarks.readthedocs.io/en/docs/umbridge/clients.html). 
    Requires [Docker](https://www.docker.com/) is installed.

    Examples:
        >>> _ = os.system('docker run --name muqbppytest -dit -p 4243:4243 linusseelinger/benchmark-muq-beam-propagation:latest > /dev/null')
        >>> import umbridge
        >>> dnb2 = DigitalNetB2(dimension=3,seed=7)
        >>> true_measure = Uniform(dnb2,lower_bound=1,upper_bound=1.05)
        >>> um_bridge_model = umbridge.HTTPModel('http://localhost:4243','forward')
        >>> um_bridge_config = {"d": dnb2.d}
        >>> integrand = UMBridgeWrapper(true_measure,um_bridge_model,um_bridge_config,parallel=False)
        >>> y = integrand(2**10)
        >>> with np.printoptions(formatter={"float":lambda x: "%.1e"%x}):
        ...     y.mean(-1)
        array([0.0e+00, 3.9e+00, 1.5e+01, 3.2e+01, 5.5e+01, 8.3e+01, 1.2e+02,
               1.5e+02, 2.0e+02, 2.4e+02, 2.9e+02, 3.4e+02, 3.9e+02, 4.3e+02,
               4.7e+02, 5.0e+02, 5.3e+02, 5.6e+02, 5.9e+02, 6.2e+02, 6.4e+02,
               6.6e+02, 6.9e+02, 7.2e+02, 7.6e+02, 7.9e+02, 8.3e+02, 8.6e+02,
               9.0e+02, 9.4e+02, 9.7e+02])
        >>> _ = os.system('docker rm -f muqbppytest > /dev/null')
        
        Custom model with independent replications
        
        >>> class TestModel(umbridge.Model):
        ...     def __init__(self):
        ...         super().__init__("forward")
        ...     def get_input_sizes(self, config):
        ...         return [1,2,3]
        ...     def get_output_sizes(self, config):
        ...         return [3,2,1]
        ...     def __call__(self, parameters, config):
        ...         out0 = [parameters[2][0],sum(parameters[2][:2]),sum(parameters[2])]
        ...         out1 = [parameters[1][0],sum(parameters[1])]
        ...         out2 = [parameters[0]]
        ...         return [out0,out1,out2]
        ...     def supports_evaluate(self):
        ...         return True
        >>> um_bridge_model = TestModel()
        >>> um_bridge_config = {}
        >>> d = sum(um_bridge_model.get_input_sizes(config=um_bridge_config))
        >>> true_measure = Uniform(DigitalNetB2(dimension=d,seed=7,replications=15),lower_bound=-1,upper_bound=1)
        >>> integrand = UMBridgeWrapper(true_measure,um_bridge_model)
        >>> y = integrand(2**6)
        >>> y.shape 
        (6, 15, 64)
        >>> muhats = y.mean(-1)
        >>> muhats.shape 
        (6, 15)
        >>> muhats_aggregate = muhats.mean(-1)
        >>> muhats_aggregate.shape 
        (6,)
        >>> muhats_agg_list_of_lists = integrand.to_umbridge_out_sizes(muhats_aggregate)
        >>> [["%.2e"%ii for ii in i] for i in muhats_agg_list_of_lists]
        [['-1.59e-08', '1.49e-04', '1.49e-04'], ['8.20e-06', '-1.38e-04'], ['-8.14e-06']]
    """

    def __init__(self, true_measure, model, config={}, parallel=False):
        """
        Args:
            true_measure (AbstractTrueMeasure): The true measure.  
            model (umbridge.HTTPModel): A `UM-Bridge` model. 
            config (dict): Configuration keyword argument to `umbridge.HTTPModel(url,name).__call__`.
            parallel (int): Parallelization flag. 
                
                - When `parallel = 0` or `parallel = 1` then function evaluation is done in serial fashion.
                - `parallel > 1` specifies the number of processes used by `multiprocessing.Pool` or `multiprocessing.pool.ThreadPool`.
            
                Setting `parallel=True` is equivalent to `parallel = os.cpu_count()`.
        """
        import umbridge
        self.parameters = []
        self.true_measure = true_measure
        self.sampler = self.true_measure 
        self.model = model
        if not self.model.supports_evaluate(): raise ParameterError("UMBridgeWrapper requires model supports evaluation.")
        self.config = config
        self.parallel = parallel
        self.d_in_umbridge = np.append(0,np.cumsum(self.model.get_input_sizes(self.config)))
        self.n_d_in_umbridge = len(self.d_in_umbridge)-1
        if self.true_measure.d!=self.d_in_umbridge[-1]:
            raise ParameterError("sampler dimension (%d) must equal the sum of UMBridgeWrapper input sizes (%d)."%(self.true_measure.d,self.d_in_umbridge[-1]))
        self.d_out_umbridge = np.append(0,np.cumsum(self.model.get_output_sizes(self.config)))
        self.n_d_out_umbridge = len(self.d_out_umbridge)-1
        self.total_out_elements = int(self.d_out_umbridge[-1])
        super(UMBridgeWrapper,self).__init__(
            dimension_indv = () if self.total_out_elements==1 else (self.total_out_elements,),
            dimension_comb = () if self.total_out_elements==1 else (self.total_out_elements,),
            parallel = self.parallel,
            threadpool = True)
    
    def g(self, t, **kwargs):
        y = np.zeros((self.total_out_elements,)+tuple(t.shape[:-1]),dtype=float)
        idxiterator = np.ndindex(t.shape[:-1])
        for i in idxiterator:
            ti_ll = [t[i+(slice(self.d_in_umbridge[j],self.d_in_umbridge[j+1]),)].tolist() for j in range(self.n_d_in_umbridge)]
            yi_ll = self.model.__call__(ti_ll,self.config)
            for j,yi_l in enumerate(yi_ll):
                y[(slice(self.d_out_umbridge[j],self.d_out_umbridge[j+1]),)+i] = yi_l if len(yi_l)>1 else yi_l[0]
        return y[0] if self.total_out_elements==1 else y
    
    def _spawn(self, level, sampler):
        return UMBridgeWrapper(
            true_measure = self.true_measure,
            model = self.model,
            config = self.config, 
            parallel = self.parallel)

    def to_umbridge_out_sizes(self, x):
        """
        Convert a data attribute to `UM-Bridge` output sized list of lists. 
        
        Args:
            x (np.ndarray): Array of length `sum(model.get_output_sizes(self.config))` where `model` is a `umbridge.HTTPModel`. 
        
        Returns:
            x_list_list (list): List of lists with sub-list lengths specified by `model.get_output_sizes(self.config)`.
        """
        return [x[...,self.d_out_umbridge[j]:self.d_out_umbridge[j+1]].tolist() for j in range(self.n_d_out_umbridge)]



import numpy as np
from .abstract_integrand import AbstractIntegrand
from ..true_measure import Uniform
from ..discrete_distribution import DigitalNetB2

class Hartmann6d(AbstractIntegrand):
    r"""
    Wrapper around [`BoTorch`'s implementation of the Augmented Hartmann function](https://botorch.readthedocs.io/en/stable/test_functions.html#botorch.test_functions.multi_fidelity.AugmentedHartmann) in dimension $d=6$. 

    Examples:
        >>> integrand = Hartmann6d(DigitalNetB2(6,seed=7))
        >>> y = integrand(2**10)
        >>> print("%.4f"%y.mean())
        -0.2644
        >>> integrand.true_measure
        Uniform (AbstractTrueMeasure)
            lower_bound     0
            upper_bound     1
            
        With independent replications

        >>> integrand = Hartmann6d(DigitalNetB2(6,seed=7,replications=2**4))
        >>> y = integrand(2**6)
        >>> y.shape
        (16, 64)
        >>> muhats = y.mean(-1) 
        >>> muhats.shape 
        (16,)
        >>> print("%.4f"%muhats.mean())
        -0.2599
    """
    
    def __init__(self, sampler):
        r"""
        Args:
            sampler (Union[AbstractDiscreteDistribution,AbstractTrueMeasure]): Either  
                
                - a discrete distribution from which to transform samples, or
                - a true measure by which to compose a transform.
        """
        self.sampler = sampler
        assert self.sampler.d==6
        self.true_measure = Uniform(self.sampler,lower_bound=0,upper_bound=1)
        super(Hartmann6d,self).__init__(dimension_indv=(),dimension_comb=(),parallel=False)
        from botorch.test_functions.multi_fidelity import AugmentedHartmann
        self.ah = AugmentedHartmann(negate=False)
        
    def g(self, t):
        import torch
        t = np.concatenate([t,np.ones(tuple(t.shape[:-1])+(1,))],axis=-1)
        return self.ah.evaluate_true(torch.tensor(t)).numpy()



from .abstract_cub_bayes_ld_g import AbstractCubBayesLDG
from ..discrete_distribution import Lattice
from ..integrand import Keister,BoxIntegral,Genz,SensitivityIndices
from ..fast_transform import fftbr,omega_fftbr
from ..util import ParameterError#, ParameterWarning #MaxSamplesWarning,
#from math import factorial
import numpy as np
#from time import time
#import warnings


class CubQMCBayesLatticeG(AbstractCubBayesLDG):
    r"""
    Quasi-Monte Carlo stopping criterion using fast Bayesian cubature and rank-1 lattices 
    with guarantees for Gaussian processes having certain shift invariant kernels.

    Examples:
        >>> k = Keister(Lattice(2, seed=123456789))
        >>> sc = CubQMCBayesLatticeG(k,abs_tol=1e-4)
        >>> solution,data = sc.integrate()
        >>> data
        Data (Data)
            solution        1.808
            comb_bound_low  1.808
            comb_bound_high 1.808
            comb_bound_diff 3.60e-05
            comb_flags      1
            n_total         2^(10)
            n               2^(10)
            time_integrate  ...
        CubQMCBayesLatticeG (AbstractStoppingCriterion)
            abs_tol         1.00e-04
            rel_tol         0
            n_init          2^(8)
            n_limit         2^(22)
            order           2^(1)
        Keister (AbstractIntegrand)
        Gaussian (AbstractTrueMeasure)
            mean            0
            covariance      2^(-1)
            decomp_type     PCA
        Lattice (AbstractLDDiscreteDistribution)
            d               2^(1)
            replications    1
            randomize       SHIFT
            gen_vec_source  kuo.lattice-33002-1024-1048576.9125.txt
            order           RADICAL INVERSE
            n_limit         2^(20)
            entropy         123456789

        Vector outputs
        
        >>> f = BoxIntegral(Lattice(3,seed=7),s=[-1,1])
        >>> abs_tol = 1e-2
        >>> sc = CubQMCBayesLatticeG(f,abs_tol=abs_tol,rel_tol=0)
        >>> solution,data = sc.integrate()
        >>> solution
        array([1.18837601, 0.95984299])
        >>> data
        Data (Data)
            solution        [1.188 0.96 ]
            comb_bound_low  [1.183 0.95 ]
            comb_bound_high [1.194 0.969]
            comb_bound_diff [0.011 0.019]
            comb_flags      [ True  True]
            n_total         2^(9)
            n               [512 256]
            time_integrate  ...
        CubQMCBayesLatticeG (AbstractStoppingCriterion)
            abs_tol         0.010
            rel_tol         0
            n_init          2^(8)
            n_limit         2^(22)
            order           2^(1)
        BoxIntegral (AbstractIntegrand)
            s               [-1  1]
        Uniform (AbstractTrueMeasure)
            lower_bound     0
            upper_bound     1
        Lattice (AbstractLDDiscreteDistribution)
            d               3
            replications    1
            randomize       SHIFT
            gen_vec_source  kuo.lattice-33002-1024-1048576.9125.txt
            order           RADICAL INVERSE
            n_limit         2^(20)
            entropy         7
        >>> sol3neg1 = -np.pi/4-1/2*np.log(2)+np.log(5+3*np.sqrt(3))
        >>> sol31 = np.sqrt(3)/4+1/2*np.log(2+np.sqrt(3))-np.pi/24
        >>> true_value = np.array([sol3neg1,sol31])
        >>> assert (abs(true_value-solution)<abs_tol).all()

        Sensitivity indices 

        >>> function = Genz(Lattice(3,seed=7))
        >>> integrand = SensitivityIndices(function)
        >>> sc = CubQMCBayesLatticeG(integrand,abs_tol=5e-2,rel_tol=0)
        >>> solution,data = sc.integrate()
        >>> data
        Data (Data)
            solution        [[0.057 0.131 0.269]
                             [0.386 0.523 0.741]]
            comb_bound_low  [[0.048 0.12  0.256]
                             [0.377 0.511 0.721]]
            comb_bound_high [[0.066 0.141 0.283]
                             [0.395 0.535 0.762]]
            comb_bound_diff [[0.018 0.021 0.027]
                             [0.018 0.024 0.04 ]]
            comb_flags      [[ True  True  True]
                             [ True  True  True]]
            n_total         2^(10)
            n               [[[1024 1024 1024]
                              [1024 1024 1024]
                              [1024 1024 1024]]
        <BLANKLINE>
                             [[1024 1024 1024]
                              [1024 1024 1024]
                              [1024 1024 1024]]]
            time_integrate  ...
        CubQMCBayesLatticeG (AbstractStoppingCriterion)
            abs_tol         0.050
            rel_tol         0
            n_init          2^(8)
            n_limit         2^(22)
            order           2^(1)
        SensitivityIndices (AbstractIntegrand)
            indices         [[ True False False]
                             [False  True False]
                             [False False  True]]
        Uniform (AbstractTrueMeasure)
            lower_bound     0
            upper_bound     1
        Lattice (AbstractLDDiscreteDistribution)
            d               3
            replications    1
            randomize       SHIFT
            gen_vec_source  kuo.lattice-33002-1024-1048576.9125.txt
            order           RADICAL INVERSE
            n_limit         2^(20)
            entropy         7

    **References:**

    1.  Jagadeeswaran, Rathinavel, and Fred J. Hickernell.  
        "Fast automatic Bayesian cubature using lattice sampling."  
        Statistics and Computing 29.6 (2019): 1215-1229.
    
    2.  Jagadeeswaran Rathinavel and Fred J. Hickernell,  
        Fast automatic Bayesian cubature using lattice sampling.  
        Stat Comput 29, 1215-1229 (2019).  
        Available from Springer [https://doi.org/10.1007/s11222-019-09895-9](https://doi.org/10.1007/s11222-019-09895-9).

    3.  Sou-Cheng T. Choi, Yuhan Ding, Fred J. Hickernell, Lan Jiang, Lluis Antoni Jimenez Rugama,  
        Da Li, Jagadeeswaran Rathinavel, Xin Tong, Kan Zhang, Yizhi Zhang, and Xuan Zhou,  
        GAIL: Guaranteed Automatic Integration Library (Version 2.3) [MATLAB Software], 2019.  
        [http://gailgithub.github.io/GAIL_Dev/](http://gailgithub.github.io/GAIL_Dev/).  
        [https://github.com/GailGithub/GAIL_Dev/blob/master/Algorithms/IntegrationExpectation/cubBayesLattice_g.m](https://github.com/GailGithub/GAIL_Dev/blob/master/Algorithms/IntegrationExpectation/cubBayesLattice_g.m).
    """

    def __init__(self, 
                 integrand, 
                 abs_tol = 1e-2, 
                 rel_tol = 0,
                 n_init = 2**8, 
                 n_limit = 2**22, 
                 error_fun="EITHER", 
                 alpha = 0.01, 
                 ptransform = 'C1SIN',
                 errbd_type = "MLE", 
                 order = 2,
                 ):
        r"""
        Args:
            integrand (AbstractIntegrand): The integrand.
            abs_tol (np.ndarray): Absolute error tolerance.
            rel_tol (np.ndarray): Relative error tolerance.
            n_init (int): Initial number of samples. 
            n_limit (int): Maximum number of samples.
            error_fun (Union[str,callable]): Function mapping the approximate solution, absolute error tolerance, and relative error tolerance to the current error bound.

                - `'EITHER'`, the default, requires the approximation error must be below either the absolue *or* relative tolerance.
                    Equivalent to setting
                    ```python
                    error_fun = lambda sv,abs_tol,rel_tol: np.maximum(abs_tol,abs(sv)*rel_tol)
                    ```
                - `'BOTH'` requires the approximation error to be below both the absolue *and* relative tolerance. 
                    Equivalent to setting
                    ```python
                    error_fun = lambda sv,abs_tol,rel_tol: np.minimum(abs_tol,abs(sv)*rel_tol)
                    ```
            alpha (np.ndarray): Uncertainty level in $(0,1)$. 
            ptransform (str): Periodization transform, see the options in [`AbstractIntegrand.f`][qmcpy.AbstractIntegrand.f].
            errbd_type (str): Options are 
                
                - `'MLE'`: Marginal Log Likelihood. 
                - `'GCV'`: Generalized Cross Validation. 
                - `'FULL'`: Full Bayes.
            order (int): Bernoulli kernel's order. If zero, choose order automatically
        """
        super(CubQMCBayesLatticeG, self).__init__(integrand, ft=fftbr, omega=omega_fftbr,
                                               ptransform=ptransform,
                                               allowed_distribs=[Lattice],
                                               kernel=self._shift_inv_kernel,
                                               abs_tol=abs_tol, rel_tol=rel_tol,
                                               n_init=n_init, n_limit=n_limit, alpha=alpha, error_fun=error_fun, errbd_type=errbd_type)
        self.order = order  # Bernoulli kernel's order. If zero, choose order automatically
        # private properties
        # full_Bayes - Full Bayes - assumes m and s^2 as hyperparameters
        # GCV - Generalized cross validation
        self.kernType = 1  # Type-1: Bernoulli polynomial based algebraic convergence, Type-2: Truncated series
        self._xfullundtype = float
        if self.discrete_distrib.order!='RADICAL INVERSE':
            raise ParameterError("CubLattice_g requires Lattice with 'RADICAL INVERSE' order")

    # Shift invariant kernel
    # C1 : first row of the covariance matrix
    # Lambda : eigen values of the covariance matrix
    # Lambda_ring = fft(C1 - 1)

    def _shift_inv_kernel(self, xun, order, theta, avoid_cancel_error, kern_type, debug_enable):
        if kern_type == 1:
            b_order = order * 2  # Bernoulli polynomial order as per the equation
            #const_mult = -(-1) ** (b_order / 2) * ((2 * np.pi) ** b_order) / factorial(b_order)
            const_mult = -(-1) ** (b_order / 2)
            if b_order == 2:
                bern_poly = lambda x: (-x * (1 - x) + 1 / 6)
            elif b_order == 4:
                bern_poly = lambda x: (((x * (1 - x)) ** 2) - 1 / 30)
            else:
                print('Error: Bernoulli order not implemented !')

            kernel_func = lambda x: bern_poly(x)
        else:
            b = order
            kernel_func = lambda x: 2 * b * (np.cos(2 * np.pi * x) - b) / (1 + b ** 2 - 2 * b * np.cos(2 * np.pi * x))
            const_mult = 1

        if avoid_cancel_error:
            # Computes C1m1 = C1 - 1
            # C1_new = 1 + C1m1 indirectly computed in the process
            (vec_C1m1, C1_alt) = self.kernel_t(theta * const_mult, kernel_func(xun))

            lambda_factor = max(abs(vec_C1m1))
            C1_alt = C1_alt / lambda_factor
            vec_C1m1 = vec_C1m1 / lambda_factor
            # eigenvalues must be real : Symmetric pos definite Kernel
            vec_lambda_ring = np.real(fftbr(vec_C1m1)*np.sqrt(vec_C1m1.shape[-1]))

            vec_lambda = vec_lambda_ring.copy()
            vec_lambda[0] = vec_lambda_ring[0] + len(vec_lambda_ring) / lambda_factor

            if debug_enable:
                # eigenvalues must be real : Symmetric pos definite Kernel
                vec_lambda_direct = np.real(fftbr(C1_alt)*np.sqrt(C1_alt.shape[-1]))  # Note: fft output unnormalized
                if sum(abs(vec_lambda_direct - vec_lambda)) > 1:
                    print('Possible error: check vec_lambda_ring computation')
        else:
            # direct approach to compute first row of the kernel Gram matrix
            vec_C1 = np.prod(1 + theta * const_mult * kernel_func(xun), 2)
            # matlab's builtin fft is much faster and accurate
            # eigenvalues must be real : Symmetric pos definite Kernel
            vec_lambda = np.real(fftbr(vec_C1)*np.sqrt(vec_C1.shape[-1]))
            vec_lambda_ring = 0
            lambda_factor = 1

        return vec_lambda, vec_lambda_ring, lambda_factor



from .abstract_stopping_criterion import AbstractStoppingCriterion
from ..util.data import Data

from ..discrete_distribution import IIDStdUniform
from ..discrete_distribution.abstract_discrete_distribution import AbstractIIDDiscreteDistribution
from ..true_measure import Gaussian, BrownianMotion, Uniform
from ..integrand import FinancialOption, Linear0, AbstractIntegrand
from ..util import MaxSamplesWarning,ParameterError
import numpy as np
from scipy.stats import norm
from time import time
import warnings


class CubMCCLT(AbstractStoppingCriterion):
    r"""
    IID Monte Carlo stopping criterion based on the Central Limit Theorem in a two step method.

    Examples:
        >>> ao = FinancialOption(IIDStdUniform(52,seed=7))
        >>> sc = CubMCCLT(ao,abs_tol=.05)
        >>> solution,data = sc.integrate()
        >>> data
        Data (Data)
            solution        1.777
            bound_low       1.723
            bound_high      1.831
            bound_diff      0.107
            n_total         74235
            time_integrate  ...
        CubMCCLT (AbstractStoppingCriterion)
            abs_tol         0.050
            rel_tol         0
            n_init          2^(10)
            n_limit         2^(30)
            inflate         1.200
            alpha           0.010
        FinancialOption (AbstractIntegrand)
            option          ASIAN
            call_put        CALL
            volatility      2^(-1)
            start_price     30
            strike_price    35
            interest_rate   0
            t_final         1
            asian_mean      ARITHMETIC
        BrownianMotion (AbstractTrueMeasure)
            time_vec        [0.019 0.038 0.058 ... 0.962 0.981 1.   ]
            drift           0
            mean            [0. 0. 0. ... 0. 0. 0.]
            covariance      [[0.019 0.019 0.019 ... 0.019 0.019 0.019]
                             [0.019 0.038 0.038 ... 0.038 0.038 0.038]
                             [0.019 0.038 0.058 ... 0.058 0.058 0.058]
                             ...
                             [0.019 0.038 0.058 ... 0.962 0.962 0.962]
                             [0.019 0.038 0.058 ... 0.962 0.981 0.981]
                             [0.019 0.038 0.058 ... 0.962 0.981 1.   ]]
            decomp_type     PCA
        IIDStdUniform (AbstractIIDDiscreteDistribution)
            d               52
            replications    1
            entropy         7
        
        Control variates 

        >>> iid = IIDStdUniform(52,seed=7)
        >>> ao = FinancialOption(iid,option="ASIAN")
        >>> eo = FinancialOption(iid,option="EUROPEAN")
        >>> lin0 = Linear0(iid)
        >>> sc = CubMCCLT(ao,abs_tol=.05,
        ...     control_variates = [eo,lin0],
        ...     control_variate_means = [eo.get_exact_value(),0])
        >>> solution,data = sc.integrate()
        >>> data
        Data (Data)
            solution        1.790
            bound_low       1.738
            bound_high      1.843
            bound_diff      0.104
            n_total         27777
            time_integrate  ...
        CubMCCLT (AbstractStoppingCriterion)
            abs_tol         0.050
            rel_tol         0
            n_init          2^(10)
            n_limit         2^(30)
            inflate         1.200
            alpha           0.010
            cv              [FinancialOption (AbstractIntegrand)
                                 option          EUROPEAN
                                 call_put        CALL
                                 volatility      2^(-1)
                                 start_price     30
                                 strike_price    35
                                 interest_rate   0
                                 t_final         1               Linear0 (AbstractIntegrand)]
            cv_mu           [4.211 0.   ]
        FinancialOption (AbstractIntegrand)
            option          ASIAN
            call_put        CALL
            volatility      2^(-1)
            start_price     30
            strike_price    35
            interest_rate   0
            t_final         1
            asian_mean      ARITHMETIC
        BrownianMotion (AbstractTrueMeasure)
            time_vec        [0.019 0.038 0.058 ... 0.962 0.981 1.   ]
            drift           0
            mean            [0. 0. 0. ... 0. 0. 0.]
            covariance      [[0.019 0.019 0.019 ... 0.019 0.019 0.019]
                             [0.019 0.038 0.038 ... 0.038 0.038 0.038]
                             [0.019 0.038 0.058 ... 0.058 0.058 0.058]
                             ...
                             [0.019 0.038 0.058 ... 0.962 0.962 0.962]
                             [0.019 0.038 0.058 ... 0.962 0.981 0.981]
                             [0.019 0.038 0.058 ... 0.962 0.981 1.   ]]
            decomp_type     PCA
        IIDStdUniform (AbstractIIDDiscreteDistribution)
            d               52
            replications    1
            entropy         7
    """

    def __init__(self, 
                 integrand, 
                 abs_tol = 1e-2, 
                 rel_tol = 0., 
                 n_init = 1024, 
                 n_limit = 2**30,
                 inflate = 1.2, 
                 alpha = 0.01, 
                 control_variates = [], 
                 control_variate_means = [],
                 ):
        r"""
        Args:
            integrand (AbstractIntegrand): The integrand.
            abs_tol (np.ndarray): Absolute error tolerance.
            rel_tol (np.ndarray): Relative error tolerance.
            n_init (int): Initial number of samples. 
            n_limit (int): Maximum number of samples.
            inflate (float): Inflation factor $\geq 1$ to multiply by the variance estimate to make it more conservative.
            alpha (np.ndarray): Uncertainty level in $(0,1)$. 
            control_variates (list): Integrands to use as control variates, each with the same underlying discrete distribution instance.
            control_variate_means (np.ndarray): Means of each control variate. 
        """
        self.parameters = ['abs_tol','rel_tol','n_init','n_limit','inflate','alpha']
        # Set Attributes
        self.abs_tol = abs_tol
        self.rel_tol = rel_tol
        self.n_init = n_init
        self.n_limit = n_limit
        assert self.n_limit>(2*self.n_init), "require n_limit is at least twic as much as n_init"
        self.alpha = alpha
        self.inflate = inflate
        assert self.inflate>=1
        assert 0<self.alpha<1
        # QMCPy Objs
        self.integrand = integrand
        self.true_measure = self.integrand.true_measure
        self.discrete_distrib = self.integrand.discrete_distrib
        super(CubMCCLT,self).__init__(allowed_distribs=[AbstractIIDDiscreteDistribution],allow_vectorized_integrals=True)
        assert self.integrand.d_indv==()
        # control variates
        self.cv_mu = np.atleast_1d(control_variate_means)
        self.cv = control_variates
        if isinstance(self.cv,AbstractIntegrand):
            self.cv = [self.cv]
            self.cv_mu = self.cv_mu[None,...]
        assert isinstance(self.cv,list), "cv must be a list of AbstractIntegrand objects"
        for cv in self.cv:
            if (not isinstance(cv,AbstractIntegrand)) or (cv.discrete_distrib!=self.discrete_distrib) or (cv.d_indv!=self.integrand.d_indv):
                raise ParameterError('''
                        Each control variates discrete distribution must be an AbstractIntegrand instance 
                        with the same discrete distribution as the main integrand. d_indv must also match 
                        that of the main integrand instance for each control variate.''')
        self.ncv = len(self.cv)
        if self.ncv>0:
            assert self.cv_mu.shape==((self.ncv,)+self.integrand.d_indv), "Control variate means should have shape (len(control variates),d_indv)."
            self.parameters += ['cv','cv_mu']
        self.z_star = -norm.ppf(self.alpha/2.)

    def integrate(self):
        t_start = time()
        data = Data(
            parameters = [
                'solution',
                'bound_low',
                'bound_high',
                'bound_diff',
                'n_total',
                'time_integrate'])
        data.xfull = np.empty((0,self.integrand.d))
        data.yfull = np.empty(0)
        if self.ncv>0:
            data.ycvfull = np.empty((self.ncv,0))
        x0 = self.discrete_distrib(n=self.n_init)
        data.xfull = np.concatenate([data.xfull,x0],0)
        y0 = self.integrand.f(x0)
        temp_a = np.maximum(np.finfo(np.float64).eps,(time()-t_start)**0.5)
        data.yfull = np.concatenate([data.yfull,y0],-1)
        if self.ncv>0:
            ycv0 = [None]*self.ncv
            for k in range(self.ncv):
                ycv0[k] = self.cv[k].f(x0)
            ycv0 = np.stack(ycv0,0)
            data.ycvfull = np.concatenate([data.ycvfull,ycv0],1)
            cvmuhats = ycv0.mean(-1)
            x4beta = ycv0-cvmuhats[:,None]
            y4beta = y0-y0.mean()
            self.beta = np.linalg.lstsq(x4beta.T,y4beta,rcond=None)[0]
            y0 = y0-((ycv0-self.cv_mu[:,None])*self.beta[:,None]).sum(0)
        data.sighat0 = y0.std(ddof=1)
        data.solution0 = y0.mean()
        temp_b = temp_a*data.sighat0
        tol_up = np.maximum(self.abs_tol,abs(data.solution0)*self.rel_tol)
        n_mu_temp = np.ceil(temp_b*(data.sighat0/temp_a)*(self.z_star*self.inflate/tol_up)**2)
        data.n_mu = int(np.maximum(self.n_init,n_mu_temp))
        if (self.n_init+data.n_mu)>self.n_limit:
            # cannot generate this many new samples
            warning_s = """
            Already generated %d samples.
            Trying to generate %d new samples would exceed n_limit = %d.
            Will instead generate %d new samples to reach n_limit.""" \
            % (int(self.n_init),int(data.n_mu),int(self.n_limit),int(self.n_limit-self.n_init))
            warnings.warn(warning_s, MaxSamplesWarning)
            data.n_mu = self.n_limit-self.n_init
        x = self.discrete_distrib(n=data.n_mu)
        data.xfull = np.concatenate([data.xfull,x],0)
        y = self.integrand.f(x)
        data.yfull = np.concatenate([data.yfull,y],-1)
        if self.ncv>0:
            ycv = [None]*self.ncv
            for k in range(self.ncv):
                ycv[k] = self.cv[k].f(x)
            ycv = np.stack(ycv,0)
            data.ycvfull = np.concatenate([data.ycvfull,ycv],1)
            y = y-((ycv-self.cv_mu[:,None])*self.beta[:,None]).sum(0)
        data.sighat = y.std(ddof=1)
        data.solution = y.mean()
        data.bound_half_width = self.z_star*self.inflate*data.sighat/np.sqrt(data.n_mu)
        data.bound_low = data.solution-data.bound_half_width
        data.bound_high = data.solution+data.bound_half_width
        data.bound_diff = data.bound_high-data.bound_low
        data.n_total = self.n_init+data.n_mu 
        data.stopping_crit = self
        data.integrand = self.integrand
        data.true_measure = self.integrand.true_measure
        data.discrete_distrib = self.true_measure.discrete_distrib
        data.time_integrate = time()-t_start
        return data.solution,data

    def set_tolerance(self, abs_tol=None, rel_tol=None, rmse_tol=None):
        assert rmse_tol is None, "rmse_tol not supported by this stopping criterion."
        if abs_tol is not None:
            self.abs_tol = abs_tol
        if rel_tol is not None:
            self.rel_tol = rel_tol
